package rbas.ranksearch.search;

import java.util.*;

import rbas.ranksearch.sem.VariationalMimMixture;
import rbas.ranksearch.sem.VariationalSemMixture;
import rbas.ranksearch.sem.MixtureOfSemsEstimator;
import edu.cmu.tetrad.data.RectangularDataSet;
import edu.cmu.tetrad.data.CovarianceMatrix;
import edu.cmu.tetrad.data.ColtDataSet;
import edu.cmu.tetrad.graph.*;

/**
 * RANKSEARCH: implements a Bayesian search for latent multiple indicator models. This is NOT officialy part of the
 * Tetrad project. I still consider this working in progress, so future developments might insert quite extra
 * features and modifications of old ones (not to mention bugs).
 *
 * This code can be quite slow. If you have a few dozen variables and a few thousand data points, expect it to take
 * several hours. I might implement extra approximations (particularly to the fastFreeVariableInsertion method).
 * If you have a large problem, I would consider setting naiveFreeVariableInsertion to true. This will include
 * all unclustered variables as children of all latents, but no extra search would be performed.
 *
 * For more information:
 *
 *  - Silva, R. and Scheines, R. (2006) "Bayesian learning of measurement and structural models". ICML
 *
 * @author Ricardo Silva, Fall 2004 - Fall 2006
 * @version $Revision: 1.1 $ $Date: 2006/09/09 19:42:51 $
 */

public class RankSearch {

    private RectangularDataSet dataSet;
    private int     numMixtures;
    private boolean naiveFreeVariableInsertion, noBidirectedEdgeSearch;

    private static final int MAXIMIZATION_TRIALS = 4;
    private static final int EXTRA_MAXIMIZATION_TRIALS = 1;

    public RankSearch(RectangularDataSet dataSet, int numMixtures) {
        this.dataSet     = dataSet;
        this.numMixtures = numMixtures;
        this.naiveFreeVariableInsertion = false;
        this.noBidirectedEdgeSearch = true;
        normalizeData();
    }

    private void normalizeData() {
        int cols = this.dataSet.getNumColumns(), rows = this.dataSet.getNumRows();
        CovarianceMatrix covMatrix = new CovarianceMatrix(this.dataSet);
        double matrix[][] = this.dataSet.getDoubleMatrix().toArray();
        double means[] = new double[this.dataSet.getNumColumns()];
        for (int i = 0; i < cols; i++)
            means[i] = 0.;
        for (int i = 0; i < rows; i++)
            for (int j = 0; j < cols; j++)
                means[j] += matrix[i][j];
        for (int j = 0; j < cols; j++)
                means[j] /= rows;
        for (int i = 0; i < rows; i++)
            for (int j = 0; j < cols; j++)
                matrix[i][j] = (matrix[i][j] - means[j]) / Math.sqrt(covMatrix.getValue(j, j));
        RectangularDataSet dataSet2 = new ColtDataSet(this.dataSet.getNumRows(), this.dataSet.getVariables());
        for (int i = 0; i < rows; i++)
            for (int j = 0; j < cols; j++)
                dataSet2.setDouble(i, j, matrix[i][j]);
        this.dataSet = dataSet2;
    }

    public Graph search() {
        Graph graph = new EdgeListGraph();
        List freeVariables = new ArrayList(); freeVariables.addAll(this.dataSet.getVariables());

        System.out.println("** ITERATED WASHDOWN");
        graph = iteratedWashdown(freeVariables);
        System.out.println("** END OF ITERATED WASHDOWN");
        System.out.println(graph); debugPrintDump(freeVariables);
        System.out.println("** FAST FREE VARIABLE INSERTION");
        graph = fastFreeVariableInsertion(graph, freeVariables);
        System.out.println("** END OF FAST FREE VARIABLE INSERTION");
        System.out.println("** LEARN STRUCTURAL MODEL");
        List latents = new ArrayList();
        for (Iterator it = graph.getNodes().iterator(); it.hasNext();) {
            Node next = (Node) it.next();
            if (next.getNodeType() == NodeType.LATENT)
                latents.add(next);
        }
        graph = greedy(graph, latents);
        System.out.println("** END OF LEARN STRUCTURAL MODEL");
        System.out.println("*** SOLUTION");
        System.out.println(graph); debugPrintDump(freeVariables);
        return graph;
    }

    /**
     * Main method: repeatedly attemp to cluster nodes by making them children of a particular set of variables.
     * This method iterates over the number of latents per cluster and makes a call to washdown for each chosen
     * number of latents.
     *
     * @return
     */

    private Graph iteratedWashdown(List freeVariables) {
        List allGraphs = new ArrayList();
        List nextFreeVariables = new ArrayList(freeVariables);
        int iter = 1, numLatents = 1;

        //Part I: get individual parts
        do {
            mainWashdownCycle:
            do {
                System.out.println("** ITERATIVE WASHDOWN - Iteration " + (iter++));
                Graph nextGraph = washdown(nextFreeVariables, numLatents);
                System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
                System.out.println(nextGraph); debugPrintDump(nextFreeVariables);
                System.out.println("&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&");
                for (Iterator it = nextGraph.getNodes().iterator(); it.hasNext();)
                    if (((Node) it.next()).getNodeType() == NodeType.LATENT) {
                        allGraphs.add(nextGraph);
                        if ((nextFreeVariables.size() * (nextFreeVariables.size() + 1)) / 2 <=
                                  numLatents * nextFreeVariables.size() + nextFreeVariables.size())
                            break;
                        else
                            continue mainWashdownCycle;
                    }
                break;
            } while (true);
            numLatents++;
            if ((nextFreeVariables.size() * (nextFreeVariables.size() + 1)) / 2 <=
                                 numLatents * nextFreeVariables.size() + nextFreeVariables.size())
                 break;
        } while (true);

        //Part II: merge graphs
        int latentCount = 0;
        Graph fullGraph = new EdgeListGraph();
        List allLatents = new ArrayList();
        for (Iterator it = allGraphs.iterator(); it.hasNext();) {
            Graph nextGraph = (Graph) it.next();
            for (Iterator it2 = nextGraph.getNodes().iterator(); it2.hasNext();) {
                Node nextNode = (Node) it2.next();
                if (nextNode.getNodeType() == NodeType.LATENT) {
                    nextNode.setName("_L" + (latentCount++));
                    allLatents.add(nextNode);
                }
                fullGraph.addNode(nextNode);
            }
            for (Iterator it2 = nextGraph.getEdges().iterator(); it2.hasNext();) {
                Edge nextEdge = (Edge) it2.next();
                if (nextEdge.getNode1().getNodeType() == NodeType.MEASURED ||
                         nextEdge.getNode2().getNodeType() == NodeType.MEASURED)
                fullGraph.addDirectedEdge(nextEdge.getNode1(), nextEdge.getNode2());
            }
        }
        for (int i = 0; i < allLatents.size() - 1; i++)
            for (int j = i + 1; j < allLatents.size(); j++)
                 fullGraph.addDirectedEdge((Node) allLatents.get(i), (Node) allLatents.get(j));
        freeVariables.clear(); freeVariables.addAll(nextFreeVariables);
        System.out.println(fullGraph.toString());

        //Part III: add impurities if necessary
        //NOTICE:
        //This is the main computational bottleneck of the procedure. I've deactivated it by default, but in principle
        //the algorithm in ICML 2006 makes use of it. This implementation is far from ideal, too. It could be
        //reasonably optimized. Activate it by setting this.noBidirectedEdgeSearch to false.

        if (this.noBidirectedEdgeSearch)
            return fullGraph;

        System.out.println("#Total number of graphs: " + allGraphs.size());
        for (int i = 0; i < allGraphs.size() - 1; i++)
             for (int j = i + 1; j < allGraphs.size(); j++) {
                 System.out.println("#Adjusting graphs " + i + " x " + j);
                 Graph partialGraph = new EdgeListGraph();
                 List partialLatents = new ArrayList();
                 List nodesToAdd = new ArrayList(((Graph) allGraphs.get(i)).getNodes());
                 nodesToAdd.addAll(((Graph) allGraphs.get(j)).getNodes());
                 for (Iterator it = nodesToAdd.iterator(); it.hasNext();) {
                      Node nextNode = (Node) it.next();
                      if (nextNode.getNodeType() == NodeType.LATENT) {
                          nextNode.setName("_L" + (latentCount++));
                          partialLatents.add(nextNode);
                      }
                      partialGraph.addNode(nextNode);
                 }
                 for (Iterator it2 = ((Graph) allGraphs.get(i)).getEdges().iterator(); it2.hasNext();) {
                     Edge nextEdge = (Edge) it2.next();
                     if (nextEdge.getNode1().getNodeType() == NodeType.MEASURED ||
                              nextEdge.getNode2().getNodeType() == NodeType.MEASURED)
                     partialGraph.addDirectedEdge(nextEdge.getNode1(), nextEdge.getNode2());
                 }
                 for (Iterator it2 = ((Graph) allGraphs.get(j)).getEdges().iterator(); it2.hasNext();) {
                     Edge nextEdge = (Edge) it2.next();
                     if (nextEdge.getNode1().getNodeType() == NodeType.MEASURED ||
                              nextEdge.getNode2().getNodeType() == NodeType.MEASURED)
                     partialGraph.addDirectedEdge(nextEdge.getNode1(), nextEdge.getNode2());
                 }
                 for (int l1 = 0; l1 < partialLatents.size() - 1; l1++)
                      for (int l2 = l1 + 1; l2 < partialLatents.size(); l2++)
                           partialGraph.addDirectedEdge((Node) partialLatents.get(l1), (Node) partialLatents.get(l2));

                 //Add impurities
                 VariationalMimMixture model = getMimGaussianModel(partialGraph, new ArrayList(), true);
                 do {
                     double currentScore = model.getScore();
                     double bestScore = -Double.MAX_VALUE;
                     Node   bestNode1 = null, bestNode2 = null;
                     for (Iterator it = ((Graph) allGraphs.get(i)).getNodes().iterator(); it.hasNext();) {
                         Node nextNode1 = (Node) it.next();
                         if (nextNode1.getNodeType() == NodeType.LATENT)
                             continue;
                         for (Iterator it2 = ((Graph) allGraphs.get(j)).getNodes().iterator(); it2.hasNext();) {
                             Node nextNode2 = (Node) it2.next();
                             if (nextNode2.getNodeType() == NodeType.LATENT)
                                 continue;
                             if (partialGraph.getEdges(nextNode1, nextNode2) != null &&
                                     partialGraph.getEdges(nextNode1, nextNode2).size() > 0)
                                 continue;
                             System.out.print("-- Adding impurity " + nextNode1.toString() + " <--> " + nextNode2.toString());
                             double newScore = model.fastBidirectedEdgeAdditionEval(nextNode1, nextNode2);
                             System.out.print(" (" + newScore + ")");
                             if (newScore > bestScore) {
                                 bestScore = newScore;
                                 bestNode1 = nextNode1; bestNode2 = nextNode2;
                                 System.out.println(" * ");
                             }
                             else
                                 System.out.println();
                         }
                     }
                     if (bestNode1 != null) {
                         Graph backupGraph = new EdgeListGraph(partialGraph);
                         backupGraph.addBidirectedEdge(bestNode1, bestNode2);
                         VariationalMimMixture model1 = getMimGaussianModel(backupGraph, new ArrayList(), true);
                         System.out.println("-- Real score = " + model1.getScore());
                         if (model1.getScore() > currentScore) {
                             System.out.println("-- Adding bidirected edge " +
                                     bestNode1.toString() + " <--> " + bestNode2.toString() +
                                     " (" + currentScore + " --> " + model1.getScore() + ")");
                             model = model1;
                             fullGraph.addBidirectedEdge(bestNode1, bestNode2);
                             partialGraph.addBidirectedEdge(bestNode1, bestNode2);
                         } else
                             break;
                     }
                     else
                          break;
                 } while (true);

                 //Remove impurities
                 Graph partialGraphAugmented = new EdgeListGraph(partialGraph);
                 List<Edge> bidirectedEdges = new ArrayList<Edge>();
                 List edgesToAdd = new ArrayList();
                 List confounders = new ArrayList();
                 for (Iterator it = partialGraphAugmented.getEdges().iterator(); it.hasNext();) {
                     Edge nextEdge = (Edge) it.next();
                     if (nextEdge.getEndpoint1() == Endpoint.ARROW && nextEdge.getEndpoint2() == Endpoint.ARROW) {
                         bidirectedEdges.add(nextEdge);
                         Node extraLatent = new GraphNode("__L" + partialGraphAugmented.getNumNodes());
                         extraLatent.setNodeType(NodeType.LATENT); partialGraphAugmented.addNode(extraLatent);
                         edgesToAdd.add(new Edge(extraLatent, nextEdge.getNode1(), Endpoint.TAIL, Endpoint.ARROW));
                         edgesToAdd.add(new Edge(extraLatent, nextEdge.getNode2(), Endpoint.TAIL, Endpoint.ARROW));
                         confounders.add(extraLatent);
                     }
                 }
                 partialGraphAugmented.removeEdges(bidirectedEdges);
                 for (Iterator it = edgesToAdd.iterator(); it.hasNext();)
                     partialGraphAugmented.addEdge((Edge) it.next());
                 model = getMimGaussianModel(partialGraphAugmented, new ArrayList(), true);
                 do {
                     double currentScore = model.getScore();
                     Node bestConfounder = null;

                     List edgesToRemove = new ArrayList();
                     double bestScore = -Double.MAX_VALUE;
                     Node bestChild1 = null, bestChild2 = null;
                     for (Iterator it = confounders.iterator(); it.hasNext();) {
                         Node confounder = (Node) it.next();
                         Node nextNode1 = (Node) partialGraphAugmented.getChildren(confounder).get(0);
                         Node nextNode2 = (Node) partialGraphAugmented.getChildren(confounder).get(1);
                         edgesToRemove.clear();
                         edgesToRemove.add(partialGraphAugmented.getEdge(confounder, nextNode1));
                         edgesToRemove.add(partialGraphAugmented.getEdge(confounder, nextNode2));
                         System.out.print("-- Trying to remove "  +
                                     nextNode1.toString() + " <--> " + nextNode2.toString());
                         double newScore = model.fastEdgesRemoval(edgesToRemove);
                         System.out.print(" (" + newScore + ")");
                         if (newScore > bestScore) {
                             bestScore = newScore;
                             bestConfounder = confounder;
                             bestChild1 = nextNode1;
                             bestChild2 = nextNode2;
                             System.out.println(" *");
                         } else
                             System.out.println();
                     }
                     if (bestChild1 != null) {
                         Graph backupGraph = new EdgeListGraph(partialGraphAugmented);
                         backupGraph.removeNode(backupGraph.getNode(bestConfounder.toString()));
                         VariationalMimMixture model1 = getMimGaussianModel(backupGraph, new ArrayList(), true);
                         System.out.println("-- Real score = " + model1.getScore());
                         if (model1.getScore() > currentScore) {
                             System.out.println("-- Removing bidirected edge " +
                                     bestChild1.toString() + " <--> " + bestChild2.toString() +
                                     " (" + currentScore + " --> " + bestScore + ")");
                             model = model1;
                             fullGraph.removeEdge(bestChild1, bestChild2);
                             partialGraphAugmented.removeNode(partialGraphAugmented.getNode(bestConfounder.toString()));
                             confounders.remove(bestConfounder);
                         } else
                             break;
                     } else
                         break;
                     System.out.println("** Remaining: " + confounders.size() + " confounders");
                 } while (true);
             }
        return fullGraph;
    }

    private Graph washdown(List freeVariables, int numLatents) {
        System.out.println("NEW WASHDOWN ITERATION");
        Graph g0 = new EdgeListGraph();
        List allVariables = new ArrayList(freeVariables);
        List bestGraphPerCycle = new ArrayList();
        List bestFreeVariablesPerCycle = new ArrayList();
        HashMap hashing = new HashMap();

        Graph g = new EdgeListGraph(g0);
        List latents = new ArrayList();
        List latentSetZero = new ArrayList();
        for (int i = 0; i < numLatents; i++) {
            Node latentZero = new GraphNode("_L" + g.getNumNodes());
            latentZero.setNodeType(NodeType.LATENT); latentSetZero.add(latentZero);
            g.addNode(latentZero);
        }
        latents.add(latentSetZero);
        int placeNew = latents.size() - 1;
        for (int f = 0; f < freeVariables.size(); f++) {
            Node newNode = new GraphNode(freeVariables.get(f).toString());
            g.addNode(newNode);
            List latentSet = (List) latents.get(placeNew);
            for (int i = 0; i < latentSet.size(); i++)
                 g.addDirectedEdge((Node) latentSet.get(i), newNode);
            hashing.put(newNode.toString(), new Integer(placeNew));
        }

        List   candidateFree = new ArrayList();
        double graphScore = scoreModel(g, candidateFree);
        double candidateScore, candidateScoreBackup;

        Graph candidateGraph = new EdgeListGraph(g);
        Graph candidateGraphBackup;
        List  candidateBackup = new ArrayList();

        Graph  bestGraphSoFar = new EdgeListGraph();
        bestGraphPerCycle.add(bestGraphSoFar);
        List   bestFreeVariablesSoFar = new ArrayList(); bestFreeVariablesSoFar.addAll(candidateFree);
        bestFreeVariablesPerCycle.add(candidateFree);
        int iter = 1;

        boolean fastMode = true;
        Hashtable originalParents = new Hashtable();

        nextBigCycle: //Build next cluster
        do {
            System.out.println("** WASHDOWN: cycle "+ (iter++));
            Node  bestChoice;
            candidateGraphBackup = new EdgeListGraph(candidateGraph);
            printEdges(candidateGraph);

            VariationalMimMixture model = getMimGaussianModel(candidateGraph, new ArrayList(), true);
            candidateScoreBackup = candidateScore = model.getScore();
            candidateBackup.clear(); candidateBackup.addAll(candidateFree);

            boolean updated = false;
            List removedNodes = new ArrayList();

            //for (int countI = 0; countI < 2; countI++) {
            for (int countI = 0; countI < 1; countI++) {
                System.out.println("****");
                System.out.println("** FIRST PART: REMOVE NODES");
                printEdges(candidateGraph); debugPrintDump(candidateFree);
                System.out.println("Next model, score0 = " + candidateScore);

                boolean deleteUpdate = false;
                do {
                    List listCandidates = new ArrayList();
                    for (Iterator it = candidateGraph.getNodes().iterator(); it.hasNext();) {
                        Node next = (Node) it.next();
                        if (next.getNodeType() == NodeType.LATENT)
                            continue;
                        if (!enoughPureSiblingsSet(next, candidateGraph, latents))
                            continue;
                        listCandidates.add(next);
                    }
                    double candidateScores[] = new double[listCandidates.size()];
                    if (listCandidates.size() == 0)
                        break;
                    double bestPossibleScore = -Double.MAX_VALUE;
                    double bestLikelihoodScore = -Double.MAX_VALUE;
                    Node bestPossibleChoice = null;
                    bestChoice = null;
                    for (int c = 0; c < listCandidates.size(); c++) {
                        Node next = (Node) listCandidates.get(c);
                        if (fastMode)
                            candidateScores[c] = model.fastNewFreeVariableScoring(next.toString());
                        else {
                            Graph otherGraph = new EdgeListGraph(candidateGraph);
                            otherGraph.removeNode(otherGraph.getNode(next.toString()));
                            otherGraph.addNode(next);
                            for (Iterator it = otherGraph.getNodes().iterator(); it.hasNext();) {
                                Node otherNode = (Node) it.next();
                                if (otherNode.getNodeType() == NodeType.MEASURED && otherNode != next)
                                    otherGraph.addDirectedEdge(otherNode, next);
                            }
                            candidateScores[c] = getMixtureOfSemsEstimator(otherGraph, false).getScore();
                        }
                        if (candidateScores[c] > bestLikelihoodScore) {
                            bestLikelihoodScore = candidateScores[c];
                            bestChoice = next;
                        }
                        if (candidateScores[c] > bestPossibleScore) {
                            bestPossibleScore = candidateScores[c];
                            bestPossibleChoice = next;
                        }
                    }
                    orderCandidates(listCandidates, candidateScores);
                    System.out.println("R0- " + model.getScore());
                    for (int i = 0; i < listCandidates.size(); i++)
                        System.out.println("R - " + listCandidates.get(i).toString() + " " + candidateScores[i]);
                    candidateScores[0] = model.fastNewFreeVariableScoring(listCandidates.get(0).toString());
                    System.out.println("Real score = " + candidateScores[0] + ", candidateScore = " + candidateScore);
                    if (candidateScores[0] < candidateScore)
                        bestChoice = null;
                    if (bestChoice != null) {
                        candidateFree.add(bestChoice); removedNodes.add(bestChoice);
                        originalParents.put(bestChoice.toString(), candidateGraph.getParents(candidateGraph.getNode(bestChoice.toString())));
                        candidateGraph.removeNode(candidateGraph.getNode(bestChoice.toString()));
                        model = getMimGaussianModel(candidateGraph, new ArrayList(), true);
                        candidateScore = model.getScore();
                        updated = true; deleteUpdate = true;
                        System.out.println("** REMOVED " + bestChoice.toString() + " (SCORE " + candidateScore + ")");
                    } else if (bestPossibleChoice != null) {
                        bestPossibleChoice = (Node) listCandidates.get(0);
                        VariationalMimMixture m2 = (VariationalMimMixture) model.clone();
                        m2.slowNewFreeVariableScoring(bestPossibleChoice.toString());
                        System.out.println("  -- slow Structural EM score = " + m2.getScore());
                        if (m2.getScore() > candidateScore) {
                            candidateFree.add(bestPossibleChoice); removedNodes.add(bestPossibleChoice);
                            originalParents.put(bestPossibleChoice.toString(), candidateGraph.getParents(candidateGraph.getNode(bestPossibleChoice.toString())));
                            candidateGraph.removeNode(candidateGraph.getNode(bestPossibleChoice.toString()));
                            model = getMimGaussianModel(candidateGraph, new ArrayList(), true);
                            candidateScore = model.getScore();
                            updated = true; deleteUpdate = true;
                            bestChoice = bestPossibleChoice;
                            System.out.println("** REMOVED " + bestPossibleChoice.toString() + " (SCORE " + candidateScore + ")");
                        } else {
                            System.out.println("-- Trying non-structural EM deletion (" + bestPossibleChoice.toString() + ")");
                            candidateFree.add(bestPossibleChoice); removedNodes.add(bestPossibleChoice);
                            Graph bestPossibleGraph = new EdgeListGraph(candidateGraph);
                            bestPossibleGraph.removeNode(bestPossibleGraph.getNode(bestPossibleChoice.toString()));;
                            List dummyFree = new ArrayList(); dummyFree.add(bestPossibleChoice);
                            VariationalMimMixture newModel = getMimGaussianModel(bestPossibleGraph, dummyFree, true);
                            if (newModel.getScore() > candidateScore) {
                                originalParents.put(bestPossibleChoice.toString(), candidateGraph.getParents(candidateGraph.getNode(bestPossibleChoice.toString())));
                                candidateGraph = bestPossibleGraph;
                                model = getMimGaussianModel(candidateGraph, new ArrayList(), true);
                                candidateScore = model.getScore();
                                System.out.println("** X-DELETED " + bestPossibleChoice.toString() + " (SCORE " + candidateScore + ")");
                                updated = true; deleteUpdate = true; bestChoice = bestPossibleChoice;
                            } else {
                                candidateFree.remove(bestPossibleChoice);
                                removedNodes.remove(bestPossibleChoice);
                                bestChoice = null;
                            }
                        }
                    }
                } while (bestChoice != null);

                if (!deleteUpdate)
                    break;
            }

            //Use as free variables only nodes that are either on bestGraphSoFar or candidateGraph,
            //but not in both

            List newlyRemoved1 = new ArrayList();
            nextBackupNode:
            for (Iterator it = bestGraphSoFar.getNodes().iterator(); it.hasNext();) {
                Node nodeB = (Node) it.next();
                if (nodeB.getNodeType() == NodeType.LATENT)
                    continue;
                for (Iterator it2 = candidateGraph.getNodes().iterator(); it2.hasNext();) {
                    Node nodeC = (Node) it2.next();
                    if (nodeB.getName().equals(nodeC.getName()))
                        continue nextBackupNode;
                }
                newlyRemoved1.add(nodeB);
            }
            System.out.println("* Scoring candidate graph with respect to the best graph so far:");
            System.out.println(candidateGraph.toString()); debugPrintDump(newlyRemoved1);
            VariationalMimMixture newModel = getMimGaussianModel(candidateGraph, newlyRemoved1, true);
            List newlyRemoved2 = new ArrayList();
            nextBackupNode2:
            for (Iterator it = candidateGraph.getNodes().iterator(); it.hasNext();) {
                Node nodeB = (Node) it.next();
                if (nodeB.getNodeType() == NodeType.LATENT)
                    continue;
                for (Iterator it2 = bestGraphSoFar.getNodes().iterator(); it2.hasNext();) {
                    Node nodeC = (Node) it2.next();
                    if (nodeB.getName().equals(nodeC.getName()))
                        continue nextBackupNode2;
                }
                newlyRemoved2.add(nodeB);
            }
            System.out.println("* Scoring best graph so far with respect to the candidate graph:");
            System.out.println(bestGraphSoFar.toString()); debugPrintDump(newlyRemoved2);
            VariationalMimMixture competingModel = getMimGaussianModel(bestGraphSoFar, newlyRemoved2, true);

            /**
             * Now, compare new model against the competing model by looking at those nodes that are new to this
             * iteration. Nodes from the previous iteration that are not here will be ignore in 'competingModel'
             */

            //Wash them down
            if (newModel == null || competingModel == null || newModel.getScore() > competingModel.getScore()) {
                System.out.println(" ** UPDATING BEST MODEL (score = " + newModel.getScore() + ")");
                bestGraphSoFar = new EdgeListGraph(candidateGraph);
                bestGraphPerCycle.add(bestGraphSoFar);
                bestFreeVariablesSoFar.clear(); bestFreeVariablesSoFar.addAll(candidateFree);
                bestFreeVariablesPerCycle.add(bestFreeVariablesSoFar);
                fastMode = true;
            } else {
                System.out.println(" ** NO SCORE IMPROVEMENT (score = " + newModel.getScore() + ")");
                break;
            }
            if (updated) {
                Graph draftGraph = new EdgeListGraph(candidateGraphBackup);
                int totalNew = 0;
                for (Iterator it = candidateFree.iterator(); it.hasNext();) {
                    Node nextFreeNode = (Node) it.next();
                    if (((Integer) hashing.get(nextFreeNode.toString())).intValue() == latents.size() - 1)
                       totalNew++;
                }
                if ((totalNew * (totalNew + 1)) / 2 >= numLatents * totalNew + totalNew) {
                    int overallNumberOfLatents = 0;
                    for (Iterator it = draftGraph.getNodes().iterator(); it.hasNext();) {
                        Node node = (Node) it.next();
                        if (node.getNodeType() == NodeType.LATENT)
                            overallNumberOfLatents++;
                    }
                    List newLatentSet = new ArrayList();
                    for (int i = 0; i < numLatents; i++) {
                        Node nextLatent = new GraphNode("_L" + overallNumberOfLatents++);
                        nextLatent.setNodeType(NodeType.LATENT);
                        draftGraph.addNode(nextLatent);
                        newLatentSet.add(nextLatent);
                        for (int j = 0; j < latents.size(); j++) {
                            List previousSet = (List) latents.get(j);
                            for (int k = 0; k < previousSet.size(); k++) {
                                Node previousLatent = draftGraph.getNode(previousSet.get(k).toString());
                                if (previousLatent != nextLatent)
                                    draftGraph.addDirectedEdge(previousLatent, nextLatent);
                            }
                        }
                    }
                    latents.add(newLatentSet);
                }
                List stillFree = new ArrayList();
                for (Iterator it = candidateFree.iterator(); it.hasNext();) {
                    Node nextFreeNode = (Node) it.next();
                    int  index = ((Integer) hashing.get(nextFreeNode.toString())).intValue();
                    if (draftGraph.getNode(nextFreeNode.toString()) != null) {
                        List latentSet = (List) latents.get(index);
                        for (int i = 0; i < latentSet.size(); i++)
                            draftGraph.removeEdge(draftGraph.getNode(latentSet.get(i).toString()),
                                                  draftGraph.getNode(nextFreeNode.toString()));
                    }
                    if (index < latents.size() - 1) {
                        if (draftGraph.getNode(nextFreeNode.toString()) == null)
                            draftGraph.addNode(new GraphNode(nextFreeNode.toString()));
                        List latentSet = (List) latents.get(index + 1);
                        for (int i = 0; i < latentSet.size(); i++)
                            draftGraph.addDirectedEdge(draftGraph.getNode(latentSet.get(i).toString()),
                                                       draftGraph.getNode(nextFreeNode.toString()));
                        hashing.remove(nextFreeNode.toString());
                        hashing.put(nextFreeNode.toString(), new Integer(index + 1));
                    }
                    else {
                        stillFree.add(nextFreeNode);
                        if (draftGraph.getNode(nextFreeNode.toString()) != null)
                            draftGraph.removeNode(draftGraph.getNode(nextFreeNode.toString()));
                    }
                }
                g = candidateGraph;
                graphScore = newModel.getScore();
                candidateFree.clear(); candidateFree.addAll(stillFree);
                candidateGraph = draftGraph;
            }
            else {
                if (candidateScoreBackup > graphScore) {
                    g = candidateGraphBackup;
                }
                break;
            }
        } while (true);

        for (int i = bestGraphPerCycle.size() - 1; i >= 0; i--) {
            bestGraphSoFar = (Graph) bestGraphPerCycle.get(i);
            if (bestGraphSoFar.getNodes().size() == 0)
                break;
            VariationalMimMixture completeModel = getMimGaussianModelSingle(bestGraphSoFar, new ArrayList(), true);
            bestFreeVariablesSoFar = (List) bestFreeVariablesPerCycle.get(i);
            System.out.println("-- Candidate model score = " + completeModel.getScore() + "(cycle " + (i + 1) + ")");
            System.out.println(bestGraphSoFar.toString());
            List variables = new ArrayList();
            for (Iterator it = bestGraphSoFar.getNodes().iterator(); it.hasNext();) {
                Node nextNode = (Node) it.next();
                if (nextNode.getNodeType() == NodeType.MEASURED)
                    variables.add(nextNode);
            }
            Graph saturated = this.getCompleteGraph(variables);
            VariationalMimMixture saturatedModel = getMimGaussianModelSingle(saturated, new ArrayList(), true);
            System.out.println("-- Saturated model score = " + saturatedModel.getScore());
            if (completeModel.getScore() > saturatedModel.getScore()) {
                System.out.println("--- New graph generated!");
                g0 = bestGraphSoFar;
                freeVariables.clear();
                nextVariable:
                for (Iterator it = allVariables.iterator(); it.hasNext();) {
                    Object next = it.next();
                    for (Iterator it2 = g0.getNodes().iterator(); it2.hasNext();) {
                        Object node = it2.next();
                        if (node.toString().equals(next.toString()))
                            continue nextVariable;
                    }
                    freeVariables.add(next);
                }
                break;
            } else if (i == 0)
                System.out.println("--- Keeping saturated!");
            else
                System.out.println("--- Trying next candidate model!");
        }
        return g0;
    }

    /**
     * Second main component of the search: all nodes that were not clustered as children of some latent are
     * introduced individually here.
     */

    private Graph fastFreeVariableInsertion(Graph g0, List freeVariables) {
        Graph basicG0 = new EdgeListGraph(g0);

        for (Iterator it = freeVariables.iterator(); it.hasNext();) {
            Object freeId = it.next();
            System.out.println("-- Inserting " + freeId.toString());

            //Candidate model: the one with the variable integrated in the main graph
            Node free = new GraphNode(freeId.toString());
            Graph g = new EdgeListGraph(basicG0); g.addNode(free);
            for (Iterator it2 = g.getNodes().iterator(); it2.hasNext();) {
                Node gNext = (Node) it2.next();
                if (gNext.getNodeType() == NodeType.LATENT)
                    g.addDirectedEdge(gNext, free);
            }
            if (!this.naiveFreeVariableInsertion)
                g = greedyIn(g, free);

            g0.addNode(free);
            for (Iterator itp = g.getParents(free).iterator(); itp.hasNext();) {
                Node parent = (Node) itp.next();
                g0.addDirectedEdge(g0.getNode(parent.toString()), free);
            }
        }
        System.out.println(g0); debugPrintDump(freeVariables);

        //Final stage: hill-climb over the freeVariables.
        System.out.println("* FINAL GREEDY SEARCH");
        if (!this.naiveFreeVariableInsertion)
            g0 = greedy(g0, freeVariables);

        return g0;
    }

    private Graph getCompleteGraph(List variables) {
        Graph fullyDirected = new EdgeListGraph(variables);
        for (int i = 0; i < fullyDirected.getNumNodes() - 1; i++)
            for (int j = i + 1; j < fullyDirected.getNumNodes(); j++)
                fullyDirected.addDirectedEdge((Node) fullyDirected.getNodes().get(i), (Node) fullyDirected.getNodes().get(j));
        return fullyDirected;
    }

    private void orderCandidates(List candidates, double scores[]) {
        List extra = new ArrayList();
        double xScores[] = new double[scores.length];
        boolean marked[] = new boolean[scores.length];
        for (int i = 0; i < marked.length; i++)
            marked[i] = false;
        int numMarked = 0;
        do {
            int pos = 0;
            while (marked[pos]) pos++;
            double bestScore = scores[pos];
            int bestPos = pos;
            for (int i = pos + 1; i < scores.length; i++) {
                if (marked[i])
                    continue;
                if (scores[i] > bestScore) {
                    bestScore = scores[i];
                    bestPos = i;
                }
            }
            extra.add(candidates.get(bestPos));
            xScores[numMarked++] = scores[bestPos];
            marked[bestPos] = true;
        } while (numMarked != candidates.size());
        candidates.clear(); candidates.addAll(extra);
        for (int i = 0; i < scores.length; i++)
            scores[i] = xScores[i];
    }

    private boolean enoughPureSiblingsSet(Node node, Graph graph, List latents) {
        for (Iterator it = latents.iterator(); it.hasNext();) {
            List latentSet = (List) it.next();
            for (Iterator it2 = latentSet.iterator(); it2.hasNext();) {
                Node latent = (Node) it2.next();
                List children = graph.getChildren(graph.getNode(latent.toString()));
                if (children.contains(graph.getNode(node.toString()))) {
                    int numIndicators = -1;
                    for (Iterator it3 = children.iterator(); it3.hasNext();) {
                        if (((Node) it3.next()).getNodeType() == NodeType.MEASURED)
                            numIndicators++;
                    }
                    if (latents.size() > 1)
                        numIndicators++;
                    return (numIndicators * (numIndicators + 1)) / 2 > numIndicators * latentSet.size() + numIndicators;
                }
            }
        }
        return false;
    }

    /*************************************************************************************************************
     * Standard search methods
     */

    private final int TABU_SIZE = 5;
    int tabuList[][] = new int[TABU_SIZE][3];
    int tabuPointer;

    /**
     * Vanilla hill-climbing search in DAGs using tabu search and Structural EM. Only edges into elements of
     * 'variables' are allowed to be modified.
     */

    private Graph greedy(Graph graph, List variables) {
        Node best_A, best_B, bestPossible_A, bestPossible_B;
        boolean improvement;
        Node nodes[] = new Node[variables.size()];
        int size = 0;
        for (Iterator it = variables.iterator(); it.hasNext();)
            nodes[size++] = graph.getNode(it.next().toString());
        boolean usedNonSEMOnce = false;
        double  bestScore = -Double.MAX_VALUE;
        VariationalSemMixture model = getMimGaussianModel(graph, new ArrayList());
        do {
            double modelScore = model.getScore();
            double bestIterationScore = -Double.MAX_VALUE;
            int type = -1, bestPossibleType = -1;
            improvement = false;
            best_A = best_B = null; bestPossible_A = bestPossible_B = null;
            for (int i = 0; i < size - 1; i++) {
                for (int j = i + 1; j < size; j++) {
                    if (!graph.isAdjacentTo(nodes[i], nodes[j])) {
                        if (!graph.existsDirectedPathFromTo(nodes[j], nodes[i])) {
                            System.out.print("* Trying to add " + nodes[i].toString() + " --> " + nodes[j].toString());
                            double newScore = modelScore + model.fastEdgeAdditionEval(nodes[i], nodes[j]);
                            if (newScore > bestScore) {
                                best_A = nodes[i];
                                best_B = nodes[j];
                                improvement = true;
                                bestScore = newScore;
                                System.out.println(" * Score improvement: " + bestScore);
                                type = 0;
                            } else {
                                System.out.println(" - score: " + newScore);
                                if (newScore > bestIterationScore) {
                                    bestPossible_A = nodes[i];
                                    bestPossible_B = nodes[j];
                                    bestPossibleType = 0;
                                    bestIterationScore = newScore;
                                }
                            }
                        }
                        if (!graph.existsDirectedPathFromTo(nodes[i], nodes[j])) {
                            System.out.print("* Trying to add " + nodes[j].toString() + " --> " + nodes[i].toString());
                            double newScore = modelScore + model.fastEdgeAdditionEval(nodes[j], nodes[i]);
                            if (newScore > bestScore) {
                                best_A = nodes[j];
                                best_B = nodes[i];
                                improvement = true;
                                bestScore = newScore;
                                System.out.println(" * Score improvement: " + bestScore);
                                type = 0;
                            } else {
                                System.out.println(" - score: " + newScore);
                                if (newScore > bestIterationScore) {
                                    bestPossible_A = nodes[j];
                                    bestPossible_B = nodes[i];
                                    bestPossibleType = 0;
                                    bestIterationScore = newScore;
                                }
                            }
                        }
                    } else {
                        Node node1, node2;
                        if (graph.isParentOf(nodes[j], nodes[i])) {
                            node1 = nodes[j];
                            node2 = nodes[i];
                        } else {
                            node1 = nodes[i];
                            node2 = nodes[j];
                        }
                        System.out.print("* Trying to remove edge " + node1.toString() + " --> " + node2.toString());
                        double newScore = modelScore + model.fastEdgeDeletionEval(node1, node2);
                        if (newScore > bestScore) {
                            best_A = node1;
                            best_B = node2;
                            improvement = true;
                            bestScore = newScore;
                            System.out.println(" * Score improvement: " + bestScore);
                            type = 1;
                        } else {
                            System.out.println(" - score: " + newScore);
                            if (newScore > bestIterationScore) {
                                bestPossible_A = node1;
                                bestPossible_B = node2;
                                bestPossibleType = 1;
                                bestIterationScore = newScore;
                            }
                        }
                        Graph copyGraph = new EdgeListGraph(graph);
                        copyGraph.removeEdges(copyGraph.getNode(node1.toString()), copyGraph.getNode(node2.toString()));
                        if (!copyGraph.existsDirectedPathFromTo(copyGraph.getNode(node1.toString()), copyGraph.getNode(node2.toString()))) {
                            System.out.print("* Trying to reverse edge " + node1.toString() + " --> " + node2.toString());
                            newScore = modelScore + model.fastEdgeReversalEval(node1, node2);
                            if (newScore > bestScore) {
                                best_A = node1;
                                best_B = node2;
                                improvement = true;
                                bestScore = newScore;
                                System.out.println(" * Score improvement: " + bestScore);
                                type = 2;
                            } else {
                                System.out.println(" - score: " + newScore);
                                if (newScore > bestIterationScore) {
                                    bestPossible_A = node1;
                                    bestPossible_B = node2;
                                    bestPossibleType = 2;
                                    bestIterationScore = newScore;
                                }
                            }
                        }
                    }
                }
            }
            if (improvement) {
                System.out.println("\n***********************************************************");
                switch (type) {
                    case 0:
                        bestScore = modelScore + model.commitFastEdgeAddition(best_A, best_B);
                        graph.addDirectedEdge(graph.getNode(best_A.toString()), graph.getNode(best_B.toString()));
                        System.out.println("* ADDING EDGE " + best_A.toString() + " --> " + best_B.toString());
                        break;
                    case 1:
                        bestScore = modelScore + model.commitFastEdgeDeletion(best_A, best_B);
                        graph.removeEdges(graph.getNode(best_A.toString()), graph.getNode(best_B.toString()));
                        System.out.println("* DELETING EDGE " + best_A.toString() + " --> " + best_B.toString());
                        break;
                    case 2:
                        bestScore = modelScore + model.commitFastEdgeReversal(best_A, best_B);
                        graph.removeEdges(graph.getNode(best_A.toString()), graph.getNode(best_B.toString()));
                        graph.addDirectedEdge(graph.getNode(best_B.toString()), graph.getNode(best_A.toString()));
                        System.out.println("* REVERSING EDGE " + best_A.toString() + " --> " + best_B.toString());
                        break;
                }
                printEdges(graph);
                System.out.println("* SCORE: " + bestScore);
                System.out.println("***********************************************************\n");
            } else if (!usedNonSEMOnce) {
                model = getMimGaussianModel(graph, new ArrayList(), true);
                improvement = true;
                System.out.println(graph.toString());
                System.out.println("* NEW STRUCTURAL EM ITERATION");
                System.out.println("* SCORE: " + bestScore);
                System.out.println("***********************************************************\n");
                usedNonSEMOnce = true;
            } else if (bestPossibleType != -1) {
                switch (bestPossibleType) {
                    case 0:
                        graph.addDirectedEdge(graph.getNode(bestPossible_A.toString()), graph.getNode(bestPossible_B.toString()));
                        break;
                    case 1:
                        graph.removeEdges(graph.getNode(bestPossible_A.toString()), graph.getNode(bestPossible_B.toString()));
                        break;
                    case 2:
                        graph.removeEdges(graph.getNode(bestPossible_A.toString()), graph.getNode(bestPossible_B.toString()));
                        graph.addDirectedEdge(graph.getNode(bestPossible_B.toString()), graph.getNode(bestPossible_A.toString()));
                        break;
                }
                model = getMimGaussianModel(graph, new ArrayList(), true);
                if (model.getScore() > bestScore) {
                   switch (bestPossibleType) {
                    case 0:
                        System.out.println("* ADDING EDGE " + bestPossible_A.toString() + " --> " + bestPossible_B.toString());
                        break;
                    case 1:
                        System.out.println("* DELETING EDGE " + bestPossible_A.toString() + " --> " + bestPossible_B.toString());
                        break;
                    case 2:
                        System.out.println("* REVERSING EDGE " + bestPossible_A.toString() + " --> " + bestPossible_B.toString());
                        break;
                    }
                    bestScore = modelScore;
                    improvement = true;
                    usedNonSEMOnce = false;
                    printEdges(graph);
                    System.out.println("* SCORE: " + bestScore);
                    System.out.println("***********************************************************\n");
                } else {
                    switch (bestPossibleType) {
                    case 0:
                        graph.removeEdges(graph.getNode(bestPossible_A.toString()), graph.getNode(bestPossible_B.toString()));
                        break;
                    case 1:
                        graph.addDirectedEdge(graph.getNode(bestPossible_A.toString()), graph.getNode(bestPossible_B.toString()));
                        break;
                    case 2:
                        graph.removeEdges(graph.getNode(bestPossible_B.toString()), graph.getNode(bestPossible_A.toString()));
                        graph.addDirectedEdge(graph.getNode(bestPossible_A.toString()), graph.getNode(bestPossible_B.toString()));
                        break;
                    }
                }
            }
        } while (improvement);

        return graph;
    }

    /**
     * Special type of greedy search where only edges into a particular target node are deleted/inserted.
     */

    private Graph greedyIn(Graph graph, Node target) {
        Node best_A;
        boolean improvement;
        Node nodes[] = new Node[graph.getNodes().size()];
        int size = 0, bestI;
        for (Iterator it = graph.getNodes().iterator(); it.hasNext();) {
            Node node = (Node) it.next();
            if (node != target)
                nodes[size++] = graph.getNode(node.toString());
        }
        this.tabuPointer = -1;
        for (int i = 0; i < TABU_SIZE; i++)
            this.tabuList[i][0] = -1;
        VariationalSemMixture model = getMimGaussianModel(graph, new ArrayList(), true);
        double bestScore = model.getScore();
        int type = -1;
        boolean usedNonSEMonce = false;
        do {
            double modelScore = model.getScore();
            double bestIterationScore = -Double.MAX_VALUE;
            int    bestPossibleI = -1, bestPossibleType = -1;
            improvement = false;
            best_A = null; bestI = -1;
            for (int i = 0; i < size; i++) {
                if (!graph.isAdjacentTo(nodes[i], target)) {
                    if (!isTabu(i, size, 1)) {
                        double newScore = modelScore + model.fastEdgeAdditionEval(nodes[i], target);
                        System.out.print("* Trying to add " + nodes[i].toString() + " --> " + target.toString());
                        if (newScore > bestScore) {
                            best_A = nodes[i]; bestI = i;
                            improvement = true;
                            bestScore = newScore;
                            //System.out.print("* Trying to add " + nodes[i].toString() + " --> " + target.toString());
                            System.out.println(" * Score improvement: " + bestScore);
                            type = 0;
                        } else {
                            System.out.println(" - Score : " + newScore);
                            if (newScore > bestIterationScore) {
                                bestPossibleI = i;
                                bestPossibleType = 0;
                                bestIterationScore = newScore;
                            }
                        }
                    }
                } else {
                    if (!isTabu(i, size, 2)) {
                        double newScore = modelScore + model.fastEdgeDeletionEval(nodes[i], target);
                        System.out.print("* Trying to remove " + nodes[i].toString() + " --> " + target.toString());
                        if (newScore > bestScore) {
                            best_A = nodes[i]; bestI = i;
                            improvement = true;
                            bestScore = newScore;
                            //System.out.print("* Trying to remove edge " + nodes[i].toString() + " --> " + target.toString());
                            System.out.println(" * Score improvement: " + bestScore);
                            type = 1;
                        } else {
                            System.out.println(" - Score : " + newScore);
                            if (newScore > bestIterationScore) {
                                bestPossibleI = i;
                                bestPossibleType = 1;
                                bestIterationScore = newScore;
                            }
                        }
                    }
                }
            }
            if (improvement) {
                System.out.println("\n***********************************************************");
                switch (type) {
                    case 0:
                        graph.addDirectedEdge(graph.getNode(best_A.toString()), graph.getNode(target.toString()));
                        model.commitFastEdgeAddition(graph.getNode(best_A.toString()), graph.getNode(target.toString()));
                        System.out.println("* ADDING EDGE " + best_A.toString() + " --> " + target.toString());
                        addTabu(bestI, size, 1);
                        break;
                    case 1:
                        graph.removeEdges(graph.getNode(best_A.toString()), graph.getNode(target.toString()));
                        model.commitFastEdgeDeletion(graph.getNode(best_A.toString()), graph.getNode(target.toString()));
                        System.out.println("* DELETING EDGE " + best_A.toString() + " --> " + target.toString());
                        addTabu(bestI, size, 2);
                        break;
                }
                System.out.println(graph.toString());
                System.out.println("* SCORE: " + bestScore);
                System.out.println("***********************************************************\n");
                usedNonSEMonce = false;
            } else if (!usedNonSEMonce) {
                model = getMimGaussianModel(graph, new ArrayList(), true);
                improvement = true;
                System.out.println(graph.toString());
                System.out.println("* SCORE: " + bestScore);
                System.out.println("***********************************************************\n");
                usedNonSEMonce = true;
            } else if (bestPossibleI != -1) {
                if (bestPossibleType == 0)
                    graph.addDirectedEdge(nodes[bestPossibleI], graph.getNode(target.toString()));
                else
                    graph.removeEdge(nodes[bestPossibleI], graph.getNode(target.toString()));
                model = getMimGaussianModel(graph, new ArrayList(), true);
                if (model.getScore() > bestScore) {
                    bestScore = model.getScore();
                    improvement = true;
                    usedNonSEMonce = false;
                    addTabu(bestPossibleI, size, bestPossibleType + 1);
                } else {
                    if (bestPossibleType == 0)
                        graph.removeEdge(nodes[bestPossibleI], graph.getNode(target.toString()));
                    else
                        graph.addDirectedEdge(nodes[bestPossibleI], graph.getNode(target.toString()));
                }
            }
        } while (improvement);

        return graph;
    }

    private boolean isTabu(int i, int j, int type) {
        for (int k = 0; k < TABU_SIZE; k++) {
            switch (type) {
                case 1: //TRYING TO ADD EDGE: check if there is no edge deletion for I -> J or J -> I
                    if (this.tabuList[k][0] == i && this.tabuList[k][1] == j && this.tabuList[k][2] == 2)
                        return true;
                    if (this.tabuList[k][0] == j && this.tabuList[k][1] == i && this.tabuList[k][2] == 2)
                        return true;
                    break;
                case 2: //TRYING TO DELETE EDGE: check if there is no edge addition for I -> J or J -> I
                    if (this.tabuList[k][0] == i && this.tabuList[k][1] == j && this.tabuList[k][2] == 1)
                        return true;
                    if (this.tabuList[k][0] == j && this.tabuList[k][1] == i && this.tabuList[k][2] == 1)
                        return true;
                    break;
                case 3: //TRYING TO REVERSE EDGE: check if there is no edge reversal for I -> J or J -> I
                    if (this.tabuList[k][0] == i && this.tabuList[k][1] == j && this.tabuList[k][2] == 3)
                        return true;
                    if (this.tabuList[k][0] == j && this.tabuList[k][1] == i && this.tabuList[k][2] == 3)
                        return true;
                    break;
            }
        }
        return false;
    }

    private void addTabu(int i, int j, int type) {
        if (this.tabuPointer == TABU_SIZE - 1)
            this.tabuPointer = 0;
        else
            this.tabuPointer++;
        this.tabuList[this.tabuPointer][0] = i; this.tabuList[this.tabuPointer][1] = j;
        this.tabuList[this.tabuPointer][2] = type;
    }

    /*************************************************************************************************************
     * Score functions
     */

    private double scoreModel(Graph graph, List nodeDump) {
        double bestScore = -Double.MAX_VALUE;
        for (int v = 0; v < MAXIMIZATION_TRIALS; v++) {
            VariationalMimMixture estimator = new VariationalMimMixture(this.dataSet, graph, nodeDump, this.numMixtures);
            estimator.estimate();
            double score = estimator.getScore();
            System.out.println("     ///score = " + score);
            if (score > bestScore)
                bestScore = score;
        }
        return bestScore;
    }

    private VariationalMimMixture getMimGaussianModel(Graph graph, List nodeDump) {
        return getMimGaussianModel(graph, nodeDump, false);
    }

    private VariationalMimMixture getMimGaussianModel(Graph graph, List nodeDump, boolean extra) {
        int totalTrials = MAXIMIZATION_TRIALS;
        if (extra)
            totalTrials += EXTRA_MAXIMIZATION_TRIALS;
        double bestScore = -Double.MAX_VALUE;
        VariationalMimMixture bestModel = null;
        for (int v = 0; v < totalTrials; v++) {
            VariationalMimMixture estimator = new VariationalMimMixture(this.dataSet, graph, nodeDump, this.numMixtures);
            estimator.estimate();
            double score = estimator.getScore();
            System.out.println("     ///score = " + score);
            if (score >= bestScore) {
                bestScore = score;
                bestModel = estimator;
            }
        }
        return bestModel;
    }

    private MixtureOfSemsEstimator getMixtureOfSemsEstimator(Graph graph, boolean extra) {
        int totalTrials = MAXIMIZATION_TRIALS;
        if (extra)
            totalTrials += EXTRA_MAXIMIZATION_TRIALS;
        double bestScore = -Double.MAX_VALUE;
        Graph backupGraph = new EdgeListGraph(graph);
        MixtureOfSemsEstimator bestModel = null;
        for (int v = 0; v < totalTrials; v++) {
            MixtureOfSemsEstimator estimator = new MixtureOfSemsEstimator(this.dataSet, backupGraph, this.numMixtures);
            estimator.estimate();
            double score = estimator.logLikelihood();
            System.out.println("     ///logLikelihood= " + score);
            if (score >= bestScore) {
                bestScore = score;
                bestModel = estimator;
            }
        }
        return bestModel;
    }

    private VariationalMimMixture getMimGaussianModelSingle(Graph graph, List nodeDump, boolean extra) {
        int totalTrials = MAXIMIZATION_TRIALS;
        if (extra)
            totalTrials += EXTRA_MAXIMIZATION_TRIALS;
        double bestScore = -Double.MAX_VALUE;
        VariationalMimMixture bestModel = null;
        for (int v = 0; v < totalTrials; v++) {
            VariationalMimMixture estimator = new VariationalMimMixture(this.dataSet, graph, nodeDump, this.numMixtures);
            estimator.estimate();
            double score = estimator.getScore();
            System.out.println("     ///score = " + score);
            if (score >= bestScore) {
                bestScore = score;
                bestModel = estimator;
            }
        }
        return bestModel;
    }

    /*************************************************************************************************************
     * Miscellaneous
     */

    private void debugPrintDump(List dump) {
        System.out.println("   --  NODE DUMP");
        for (Iterator it = dump.iterator(); it.hasNext();)
            System.out.println("           " + it.next().toString());
        System.out.println();
    }

    private void printEdges(Graph graph) {
        for (Iterator it = graph.getEdges().iterator(); it.hasNext();) {
            Edge next = (Edge) it.next();
            if (next.getNode1().getNodeType() != NodeType.LATENT || next.getNode2().getNodeType() != NodeType.LATENT)
               System.out.println("      ** " + next.toString());
        }
    }

}


