
package rbas.ranksearch.sem;

import edu.cmu.tetrad.util.MatrixUtils;
import edu.cmu.tetrad.graph.*;
import edu.cmu.tetrad.data.RectangularDataSet;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Random;

import rbas.util.RBASProbUtils;

/**
 * Implements a variational approximation for a mixture of Gaussian SEMs with special Structural EM-like evaluations
 * for multiple indicator models. This extends the VariationalSemMixture class.
 *
 * For more information:
 *
 *  - Silva, R. and Scheines, R. (2006) "Bayesian learning of measurement and structural models". ICML
 *  - Silva, R. (2005) Automatic Discovery of Latent Variable Models, PhD Thesis, Machine Learning Department, CMU
 *
 * @author Ricardo Silva, Fall 2004 - Fall 2006
 * @version $Revision: 1.1 $ $Date: 2006/09/09 19:42:51 $
 */

public class VariationalMimMixture extends VariationalSemMixture {

    public VariationalMimMixture(RectangularDataSet dataSet, Graph graph, List freeVariables, int numMixtures) {
        super(dataSet, graph, freeVariables, numMixtures);
        this.newLikelihoodKernels = new double[this.numMixtures][this.sampleSize];
    }

    public void estimateFromLastPoint() {
        double f, newF = fEval();
        int iter = 0;
        this.converging = true;
        do {
            f = newF;
            updateQUpsilon();
            updateQPi();
            updateQLambda(); //updateLambdaHyperParameters();
            updateQBetaMu(); //updateBetaMuHyperParameters();
            updateQX();
            updateQPsi(); //updatePsiHyperParameters();
            updateQS();
            updateLambdaHyperParameters(); updateBetaMuHyperParameters();
            newF = fEval();
            //System.out.println(" -- lastPoint = " + newF);
            iter++;
        } while (this.converging && newF - f > MIN_TOLERANCE && iter < MAX_ITER);
        if (!this.converging || newF < f)
            this.score = -Double.MAX_VALUE;
        else
            this.score = newF;
    }

    public void estimateFromLastPoint1() {
        this.converging = true;
        double f, newF = fEval();
        int iter = 0;
        do {
            f = newF;
            updateQUpsilon();
            newF = fEval(); System.out.println(" ------ F (Upsilon) = " + newF);
            updateQPi();
            newF = fEval(); System.out.println(" ------ F (Pi)      = " + newF);
            updateQLambda(); updateLambdaHyperParameters();
            newF = fEval(); System.out.println(" ------ F (Lambda)  = " + newF);
            updateQBetaMu(); updateBetaMuHyperParameters();
            newF = fEval(); System.out.println(" ------ F (Beta)    = " + newF);
            updateQX();
            newF = fEval(); System.out.println(" ------ F (X)       = " + newF);
            updateQPsi();
            newF = fEval(); System.out.println(" ------ F (Psi)     = " + newF);
            updateQS();
            newF = fEval(); System.out.println(" ------ F (S)       = " + newF);
            newF = fEval();
            System.out.println(" ------ lastPoint = " + newF);
            iter++;
        } while (this.converging && newF - f > MIN_TOLERANCE && iter < MAX_ITER);
        if (!this.converging || newF < f)
            this.score = -Double.MAX_VALUE;
        else
            this.score = newF;
    }

    /****************************************************************************************************************
     * STRUCTURAL EM METHODS
     */

    double newQLambdaCov[][][], newQLambdaMean[][], newQELambda[][], newQELambdaLambda[][][];
    double newQEInvPsi, newQInvPsiA, newQInvPsiB;
    int    newYYParents[], newYXParents[];
    double newLambdaMeanPrior[][], newLambdaInvCovPrior[][][], newLambdaCovPrior[][][];
    double newLikelihoodKernels[][];

    /**
     * This method rapidly re-evaluates the score function if some 'variable' is treated
     * as a 'FreeVariable'. It assumes the given 'variable' has no children (as it would be
     * the case in the main variables of washdown, in class MimScoreSearch). This is motivated
     * by Structural EM as a method for speeding up search.
     * 
     * @param variable
     * @return
     */

    public double commitFastNewFreeVariableScoring(String variable) {
        return fastNewFreeVariableScoring(variable, true);
    }

    public double fastNewFreeVariableScoring(String variable) {
        return fastNewFreeVariableScoring(variable, false);
    }

    private double fastNewFreeVariableScoring(String variable, boolean commit) {
        boolean yIn[] = new boolean[this.numVars];

        int vy = -1;
        for (int i = 0; i < this.numVars; i++) {
            yIn[i] = false;
            if (this.indicators[i].toString().equals(variable)) {
                vy = i;
                break;
            }
        }
        if (vy < 0)
            throw new RuntimeException("This variable does not exist: " + variable);

        Node nodeY = this.graph.getNode(this.indicators[vy].toString());
        List modifiedVariables = new ArrayList(); modifiedVariables.add(nodeY); yIn[vy] = true;
        for (int i = 0; i < this.numVars; i++) {
            Node next = this.graph.getNode(this.indicators[i].toString());
            if (this.graph.isParentOf(nodeY, next)) {
                modifiedVariables.add(next);
                yIn[i] = true;
            }
        }
        Graph newGraph = new EdgeListGraph(this.graph);
        newGraph.removeNode(newGraph.getNode(nodeY.toString()));
        newGraph.addNode(nodeY);
        for (Iterator it = newGraph.getNodes().iterator(); it.hasNext();) {
            Node next = (Node) it.next();
            if (next.getNodeType() == NodeType.MEASURED && next != nodeY)
                newGraph.addDirectedEdge(next, nodeY);
        }

        double totalNewExtraScore = 0.;
        updateSingleQX(vy);
        for (int k = 0; k < modifiedVariables.size(); k++) {
            String nextVariable = modifiedVariables.get(k).toString();
            int y = -1;
            for (int i = 0; i < this.numVars; i++) {
                if (this.indicators[i].toString().equals(nextVariable)) {
                    y = i;
                    break;
                }
            }
            //Estimate its new parameters
            List parents = newGraph.getParents(newGraph.getNode(this.indicators[y].toString()));
            this.newYYParents = new int[parents.size() + 1];
            for (int i = 0; i < parents.size(); i++) {
                int p = -1;
                for (int j = 0; j < this.numVars; j++)
                    if (this.indicators[j].toString().equals(parents.get(i).toString())) {
                        p = j;
                        break;
                    }
                this.newYYParents[i] = p;
            }
            this.newYYParents[parents.size()] = this.numVars;
            this.newYXParents = new int[0];
            totalNewExtraScore += singleOptimize(y);
            if (commit) {
                this.yyParents[y] = this.newYYParents;
                this.yxParents[y] = this.newYXParents;
                this.qInvPsiA[y] = this.newQInvPsiA;
                this.qInvPsiB[y] = this.newQInvPsiB;
                this.qEInvPsi[y] = this.newQEInvPsi;
                for (int d = 0; d < this.numMixtures; d++) {
                    this.qLambdaCov[d][y] = this.newLambdaCovPrior[d];
                    this.qELambdaLambda[d][y] = this.newQELambdaLambda[d];
                    this.qLambdaMean[d][y] = this.newQLambdaMean[d];
                    this.qELambda[d][y] = this.newQELambda[d];
                    this.lambdaMeanPrior[d][y] = this.newLambdaMeanPrior[d];
                    this.lambdaInvCovPrior[d][y] = this.newLambdaInvCovPrior[d];
                    this.lambdaCovPrior[d][y] = this.newLambdaCovPrior[d];
                    for (int i = 0; i < this.sampleSize; i++)
                        this.likelihoodKernels[d][i][y] = this.newLikelihoodKernels[d][i];
                }
            }
        }

        double currentScore = -RBASProbUtils.klDivergenceDirichlet(this.qAlphaM, this.alphaMStar);
        double alphaMTotal = 0;
        for (int i = 0; i < this.qAlphaM.length; i++)
            alphaMTotal += this.qAlphaM[i];
        for (int d = 0; d < this.numMixtures; d++) {
            for (int l = 0; l < this.numLatents; l++) {
                if (this.xxParents[l].length == 0)
                    continue;
                currentScore -= RBASProbUtils.klDivergenceNormal(this.qBetaMuMean[d][l], this.qBetaMuCov[d][l],
                                                             this.betaMuMeanPrior[d][l], this.betaMuCovPrior[d][l]);
            }
            for (int y = 0; y < this.numVars; y++) {
                if (!yIn[y]) {
                    currentScore -= RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][y], this.qLambdaCov[d][y],
                                                                     this.lambdaMeanPrior[d][y], this.lambdaCovPrior[d][y]);
                }
            }
            for (int i = 0; i < this.sampleSize; i++) {
                if (this.qS[d][i] == 0.)
                    continue;
                double partial = RBASProbUtils.digamma(this.qAlphaM[d]) - RBASProbUtils.digamma(alphaMTotal)
                                 - Math.log(this.qS[d][i]);
                if (this.numLatents > 0)
                    partial += expectedLnPX(i, d) + RBASProbUtils.gaussianEntropy(this.qXCov[d][i]);
                for (int y = 0; y < this.numVars; y++)
                    if (!yIn[y])
                        partial -= 0.5 * this.qEInvPsi[y] * this.likelihoodKernels[d][i][y];
                currentScore += this.qS[d][i] * partial;
                if (Double.isNaN(currentScore))
                    System.out.println("NaN Here 3");
            }
        }
        double logDetPsi = 0.;
        for (int y = 0; y < this.numVars; y++)
            if (!yIn[y])
                //logDetPsi += RBASProbUtils.digamma(this.qInvPsiA[y]) - Math.log(this.qInvPsiB[y]);
                logDetPsi += Math.log(this.qEInvPsi[y]);
        currentScore += 0.5 * logDetPsi * this.sampleSize;

        if (commit) {
            this.freeVariables.add(this.indicators[vy]);
            this.graph = newGraph;
            this.originalGraph.removeNode(this.originalGraph.getNode(this.indicators[vy].toString()));
            this.score = currentScore + totalNewExtraScore;
        } else
            updateQX();
        return currentScore + totalNewExtraScore;
    }

    public double slowNewFreeVariableScoring(String variable) {
        boolean yIn[] = new boolean[this.numVars];

        int vy = -1;
        for (int i = 0; i < this.numVars; i++) {
            yIn[i] = false;
            if (this.indicators[i].toString().equals(variable)) {
                vy = i;
                break;
            }
        }
        if (vy < 0)
            throw new RuntimeException("This variable does not exist: " + variable);

        Node nodeY = this.graph.getNode(this.indicators[vy].toString());
        List modifiedVariables = new ArrayList(); modifiedVariables.add(nodeY); yIn[vy] = true;
        for (int i = 0; i < this.numVars; i++) {
            Node next = this.graph.getNode(this.indicators[i].toString());
            if (this.graph.isParentOf(nodeY, next)) {
                modifiedVariables.add(next);
                yIn[i] = true;
            }
        }

        Graph newGraph = new EdgeListGraph(this.graph);
        newGraph.removeNode(newGraph.getNode(nodeY.toString()));
        newGraph.addNode(nodeY);
        for (Iterator it = newGraph.getNodes().iterator(); it.hasNext();) {
            Node next = (Node) it.next();
            if (next.getNodeType() == NodeType.MEASURED && next != nodeY)
                newGraph.addDirectedEdge(next, nodeY);
        }
        this.graph = newGraph;
        this.freeVariables.add(nodeY);

        Random r = new Random();
        for (int y = 0; y < this.numVars; y++) {
            if (!yIn[y])
                continue;
            List parents = newGraph.getParents(newGraph.getNode(this.indicators[y].toString()));
            this.yyParents[y] = new int[parents.size() + 1];
            for (int i = 0; i < parents.size(); i++) {
                int p = -1;
                for (int j = 0; j < this.numVars; j++)
                    if (this.indicators[j].toString().equals(parents.get(i).toString())) {
                        p = j;
                        break;
                    }
                this.yyParents[y][i] = p;
            }
            this.yyParents[y][parents.size()] = this.numVars;
            this.yxParents[y] = new int[0];
            int numParents = this.yyParents[y].length + this.yxParents[y].length;
            for (int d = 0; d < this.numMixtures; d++) {
                this.lambdaMeanPrior[d][y] = new double[numParents];
                this.lambdaInvCovPrior[d][y] = new double[numParents][numParents];
                this.lambdaCovPrior[d][y] = new double[numParents][numParents];
                this.qLambdaMean[d][y] = new double[numParents];
                this.qLambdaCov[d][y] = new double[numParents][numParents];
                this.qELambda[d][y] = new double[numParents];
                this.qELambdaLambda[d][y] = new double[numParents][numParents];
                for (int v1 = 0; v1 < numParents; v1++) {
                    this.lambdaMeanPrior[d][y][v1] = 0.; //This *has* to be zero.
                    for (int v2 = 0; v2 < numParents; v2++) {
                        if (v1 == v2) {
                            this.lambdaInvCovPrior[d][y][v1][v2] = this.lambdaInvCovPriorParam[y];
                            this.lambdaCovPrior[d][y][v1][v2] = 1. / this.lambdaInvCovPriorParam[y];
                        }
                        else {
                            this.lambdaInvCovPrior[d][y][v1][v2] = 0.;
                            this.lambdaCovPrior[d][y][v1][v2] = 0.;
                        }
                    }
                }
                for (int p = 0; p < this.qLambdaMean[d][y].length; p++) {
                    this.qLambdaMean[d][y][p] = r.nextGaussian();
                    for (int p2 = 0; p2 < this.qLambdaMean[d][y].length; p2++) {
                        if (p == p2)
                            this.qLambdaCov[d][y][p][p2] = r.nextDouble();
                        else
                            this.qLambdaCov[d][y][p][p2] = 0.;
                    }
                }
                for (int p = 0; p < this.qLambdaMean[d][y].length; p++)
                    for (int p2 = 0; p2 < this.qLambdaMean[d][y].length; p2++)
                        this.qELambdaLambda[d][y][p][p2] = this.qLambdaCov[d][y][p][p2] +
                                                           this.qELambda[d][y][p] * this.qELambda[d][y][p2];
            }
            this.qInvPsiA[y] = r.nextDouble(); this.qInvPsiB[y] = r.nextDouble();
            this.qEInvPsi[y] = this.qInvPsiA[y] / this.qInvPsiB[y];
        }
        updateQX();

        estimateFromLastPoint();
        return this.score;
    }

    private List getEndpoints(List edges) {
        List nodes = new ArrayList();
        for (Iterator it = edges.iterator(); it.hasNext();) {
            Edge nextEdge = (Edge) it.next();
            Node node1 = this.graph.getNode(nextEdge.getNode1().toString());
            Node node2 = this.graph.getNode(nextEdge.getNode2().toString());
            if (!nodes.contains(node2))
                nodes.add(node2); //only the head of the edge
            if (nextEdge.getEndpoint1() == Endpoint.ARROW && nextEdge.getEndpoint2() == Endpoint.ARROW && !nodes.contains(node1))
                nodes.add(node1); //both heads are added here
        }
        return nodes;
    }

    private List getEdgeParents(String node, List edges) {
        List parents = new ArrayList();
        for (Iterator it = edges.iterator(); it.hasNext();) {
            Edge nextEdge = (Edge) it.next();
            Node node1 = this.graph.getNode(nextEdge.getNode1().toString());
            Node node2 = this.graph.getNode(nextEdge.getNode2().toString());
            if (nextEdge.getEndpoint1() == Endpoint.ARROW && nextEdge.getEndpoint2() == Endpoint.ARROW )
                continue;
            else if (node2.toString().equals(node) && !parents.contains(node1))
                parents.add(node1);
        }
        return parents;
    }

    private List getEdgeConfounders(String node, List edges) {
        List parents = new ArrayList();
        for (Iterator it = edges.iterator(); it.hasNext();) {
            Edge nextEdge = (Edge) it.next();
            Node node1 = this.graph.getNode(nextEdge.getNode1().toString());
            Node node2 = this.graph.getNode(nextEdge.getNode2().toString());
            if (nextEdge.getEndpoint1() == Endpoint.ARROW && nextEdge.getEndpoint2() == Endpoint.ARROW ) {
                if (!node1.toString().equals(node) && !node2.toString().equals(node))
                    continue;
                for (int i = 0; i < this.confounders.length; i++) {
                    Node confounder1 = this.graph.getNode(this.confounders[i].getNode1().toString());
                    Node confounder2 = this.graph.getNode(this.confounders[i].getNode2().toString());
                    if ((confounder1 == node1 && confounder2 == node2) ||
                            (confounder1 == node2 && confounder2 == node1)) {
                        if (!parents.contains(this.confounders[i]))
                            parents.add(this.confounders[i]);
                    }
                }
            }
        }
        return parents;
    }

    private List getConfounders(Node node) {
        List confounders = new ArrayList();
        for (int i = 0; i < this.confounders.length; i++) {
            Node confounder1 = this.graph.getNode(this.confounders[i].getNode1().toString());
            Node confounder2 = this.graph.getNode(this.confounders[i].getNode2().toString());
            if (confounder1 == node || confounder2 == node)
                confounders.add(this.confounders[i]);
        }
        return confounders;
    }

    public double fastEdgesRemoval(List edges) {
        List variables = getEndpoints(edges);
        double currentScore = this.score;
        int ys[] = new int[variables.size()];
        for (int v = 0; v < variables.size(); v++) {
            ys[v] = -1;
            for (int i = 0; i < this.numVars; i++) {
                if (this.indicators[i].toString().equals(variables.get(v).toString())) {
                    ys[v] = i;
                    break;
                }
            }
            if (ys[v] < 0)
                throw new RuntimeException("This variable does not exist: " + variables.get(v));
        }
        //Remove its penalty contribution
        //for (int v = 0; v < variables.size(); v++)
        //    currentScore += RBASProbUtils.klDivergenceGamma(this.qInvPsiA[ys[v]], this.qInvPsiB[ys[v]], this.invPsiA, this.invPsiB);
        for (int d = 0; d < this.numMixtures; d++)
            for (int v = 0; v < variables.size(); v++) {
                //Remove more of its penalty contribution
                currentScore += RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][ys[v]], this.qLambdaCov[d][ys[v]],
                                                             this.lambdaMeanPrior[d][ys[v]], this.lambdaCovPrior[d][ys[v]]);
                //Remove its likelihood contribution
                for (int i = 0; i < this.sampleSize; i++)
                    currentScore += this.qS[d][i] * 0.5 * this.qEInvPsi[ys[v]] * this.likelihoodKernels[d][i][ys[v]];
            }
        //Remove more of its likelihood contribution
        for (int v = 0; v < variables.size(); v++)
            currentScore -= 0.5 * Math.log(this.qEInvPsi[ys[v]]) * this.sampleSize;
        //Estimate its new parameters
        for (int v = 0; v < variables.size(); v++) {
            List parentList = this.graph.getParents(this.graph.getNode(variables.get(v).toString()));
            List confounderList = getConfounders(this.graph.getNode(variables.get(v).toString()));
            List parentForbidden = getEdgeParents(variables.get(v).toString(), edges);
            List confounderForbidden = getEdgeConfounders(variables.get(v).toString(), edges);
            List yParentsList = new ArrayList();
            List xParentsList = new ArrayList();
            List cParentsList = new ArrayList();
            for (Iterator it = parentList.iterator(); it.hasNext();) {
                Node parent = (Node) it.next();
                if (!parentForbidden.contains(parent)) {
                    if (parent.getNodeType() == NodeType.LATENT)
                        xParentsList.add(parent);
                    else
                        yParentsList.add(parent);
                }
            }
            for (Iterator it = confounderList.iterator(); it.hasNext();) {
                Edge confounder = (Edge) it.next();
                if (!confounderForbidden.contains(confounder)) {
                    cParentsList.add(confounder);
                }
            }
            this.newYYParents = new int[yParentsList.size() + 1];
            for (int i = 0; i < yParentsList.size(); i++)
                for (int j = 0; j < this.indicators.length; j++)
                    if (this.indicators[j].toString().equals(yParentsList.get(i).toString())) {
                        this.newYYParents[i] = j;
                        break;                                                  
                    }
            this.newYYParents[yParentsList.size()] = this.numVars;
            this.newYXParents = new int[xParentsList.size() + cParentsList.size()];
            for (int i = 0; i < xParentsList.size(); i++)
                for (int j = 0; j < this.latents.length; j++) {
                    if (this.latents[j].toString().equals(xParentsList.get(i).toString())) {
                        this.newYXParents[i] = j;
                        break;
                    }
                }
            for (int i = 0; i < cParentsList.size(); i++)
                for (int j = 0; j < this.confounders.length; j++)
                    if (this.confounders[j].toString().equals(cParentsList.get(i).toString())) {
                        this.newYXParents[xParentsList.size() + i] = this.numMainLatents + j;
                        break;
                    }
            currentScore += singleOptimize(ys[v]);
        }
        return currentScore;
    }

    private void initializeNewParams(int y) {
        Random r = new Random();
        int numParents = this.newYYParents.length + this.newYXParents.length;

        this.newQLambdaCov = new double[this.numMixtures][numParents][numParents];
        this.newQELambdaLambda = new double[this.numMixtures][numParents][numParents];
        this.newQLambdaMean = new double[this.numMixtures][numParents];
        this.newQELambda = new double[this.numMixtures][numParents];
        this.newLambdaMeanPrior = new double[this.numMixtures][numParents];
        this.newLambdaInvCovPrior = new double[this.numMixtures][numParents][numParents];
        this.newLambdaCovPrior = new double[this.numMixtures][numParents][numParents];

        for (int d = 0; d < this.numMixtures; d++) {
            for (int v1 = 0; v1 < numParents; v1++) {
                this.newLambdaMeanPrior[d][v1] = 0.;
                for (int v2 = 0; v2 < numParents; v2++) {
                    if (v1 == v2) {
                        this.newLambdaInvCovPrior[d][v1][v2] = this.lambdaInvCovPriorParam[v1];
                        this.newLambdaCovPrior[d][v1][v2] = 1. / this.lambdaInvCovPriorParam[v1];
                    }
                    else {
                        this.newLambdaInvCovPrior[d][v1][v2] = 0.;
                        this.newLambdaCovPrior[d][v1][v2] = 0.;
                    }
                }
            }
            for (int p = 0; p < this.newQLambdaMean[d].length; p++) {
                this.newQLambdaMean[d][p] = r.nextGaussian();
                for (int p2 = 0; p2 < this.newQLambdaMean[d].length; p2++) {
                    if (p == p2)
                        this.newQLambdaCov[d][p][p2] = 1.;
                    else
                        this.newQLambdaCov[d][p][p2] = 0.;
                }
            }
            for (int p = 0; p < this.newQLambdaMean[d].length; p++)
                for (int p2 = 0; p2 < this.newQLambdaMean[d].length; p2++)
                    this.newQELambdaLambda[d][p][p2] = this.newQLambdaCov[d][p][p2] +
                                                       this.newQELambda[d][p] * this.newQELambda[d][p2];
        }
        this.newQInvPsiA = r.nextDouble(); this.newQInvPsiB = r.nextDouble();
        this.newQEInvPsi = this.newQInvPsiA / this.newQInvPsiB;
        computeNewLikelihoodKernels(y);
    }

    /**
     * Update the variational distribution of the edges (and intercept) into an observed variable y
     */

    private double singleOptimize(int y) {
        initializeNewParams(y);
        double f, newF = newSingleFEval();
        int iter = 0;
        do {
            f = newF;
            updateSingleQLambda(y);
            updateSingleQPsi();
            newF = newSingleFEval();
            iter++;
        } while (newF - f > MIN_TOLERANCE && iter < 5);
        return newF;
    }

    private double newSingleFEval() {
        double f = 0.;

        //f -= RBASProbUtils.klDivergenceGamma(this.newQInvPsiA, this.newQInvPsiB, this.invPsiA, this.invPsiB);
        for (int d = 0; d < this.numMixtures; d++) {
            f -= RBASProbUtils.klDivergenceNormal(this.newQLambdaMean[d], this.newQLambdaCov[d],
                                              this.newLambdaMeanPrior[d], this.newLambdaCovPrior[d]);
            for (int i = 0; i < this.sampleSize; i++)
                f -= 0.5 * this.qS[d][i] * this.newQEInvPsi* this.newLikelihoodKernels[d][i];
        }
        //f += 0.5 * (RBASProbUtils.digamma(newQInvPsiA) - Math.log(this.newQInvPsiB))  * this.sampleSize;
        f += 0.5 * Math.log(this.newQEInvPsi) * this.sampleSize;
        return f;
    }

    protected void updateSingleQLambda(int y) {
        int totalParents = this.newYYParents.length + this.newYXParents.length;
        for (int d = 0; d < this.numMixtures; d++) {
            double auxS[][] = new double[totalParents][totalParents];
            double auxMean[] = new double[totalParents];
            for (int i = 0; i < auxS.length; i++) {
                auxMean[i] = 0.;
                for (int j = 0; j < auxS.length; j++)
                    auxS[i][j] = 0.;
            }
            for (int i = 0; i < this.sampleSize; i++) {
                double constant = this.qS[d][i] * this.newQEInvPsi, nextTerm;
                for (int p1 = 0; p1 < this.newYYParents.length; p1++) {
                    for (int p2 = 0; p2 < this.newYYParents.length; p2++)
                        auxS[p1][p2] += constant *
                                        this.rawData[i][this.newYYParents[p1]] * this.rawData[i][this.newYYParents[p2]];
                    for (int p2 = 0; p2 < this.newYXParents.length; p2++) {
                        nextTerm = constant *
                                   this.rawData[i][this.newYYParents[p1]] * this.qEX[d][i][this.newYXParents[p2]];
                        auxS[p1][this.newYYParents.length + p2] += nextTerm;
                        auxS[this.newYYParents.length + p2][p1] += nextTerm;
                    }
                    auxMean[p1] += constant * this.rawData[i][y] * this.rawData[i][this.newYYParents[p1]];
                }
                for (int p1 = 0; p1 < this.newYXParents.length; p1++) {
                    for (int p2 = 0; p2 < this.newYXParents.length; p2++)
                        auxS[this.newYYParents.length + p1][this.newYYParents.length + p2] += constant *
                                this.qEXX[d][i][this.newYXParents[p1]][this.newYXParents[p2]];
                    auxMean[this.newYYParents.length + p1] +=
                            constant * this.rawData[i][y] * this.qEX[d][i][this.newYXParents[p1]];
                }
            }
            this.newQLambdaCov[d] = MatrixUtils.inverse(MatrixUtils.sum(auxS, this.newLambdaInvCovPrior[d]));
            this.newQLambdaMean[d] = MatrixUtils.product(this.newQLambdaCov[d], auxMean);
            this.newQELambda[d] = this.newQLambdaMean[d];
            for (int v1 = 0; v1 < totalParents; v1++)
                for (int v2 = 0; v2 < totalParents; v2++)
                    this.newQELambdaLambda[d][v1][v2] = this.newQLambdaCov[d][v1][v2] +
                                                        this.newQLambdaMean[d][v1] * this.newQLambdaMean[d][v2];
        }
        computeNewLikelihoodKernels(y);
    }

    protected void updateSingleQPsi() {
        double dataContrib = 0.;
        for (int d = 0; d < this.numMixtures; d++)
            for (int i = 0; i < this.sampleSize; i++)
                dataContrib += this.qS[d][i] * this.newLikelihoodKernels[d][i];
        //this.newQInvPsiA = this.invPsiA + 0.5 * this.sampleSize;
        //this.newQInvPsiB = this.invPsiB + 0.5 * dataContrib;
        //this.newQEInvPsi = this.newQInvPsiA / this.newQInvPsiB;
        this.newQEInvPsi = this.sampleSize / dataContrib;
    }

    protected void updateSingleQX(int y0) {
        if (this.numLatents == 0)
            return;
        double auxS[][] = new double[this.numLatents][this.numLatents];
        double auxM[] = new double[this.numLatents];
        for (int d = 0; d < this.numMixtures; d++) {
            double priorIntercept[] = MatrixUtils.product(this.xInvCovPrior[d], this.xMeanPrior[d]);
            for (int i = 0; i < this.sampleSize; i++) {
                for (int x = 0; x < this.numLatents; x++)
                    auxM[x] = 0.;
                for (int v1 = 0; v1 < this.numLatents; v1++)
                    for (int v2 = 0; v2 < this.numLatents; v2++)
                        auxS[v1][v2] = 0.;
                for (int y = 0; y < this.numVars; y++) {
                    if (y == y0)
                        continue;
                    for (int c1 = 0; c1 < this.yxParents[y].length; c1++) {
                        auxM[this.yxParents[y][c1]] += this.qEInvPsi[y] * this.rawData[i][y] *
                                                       this.qELambda[d][y][this.yyParents[y].length + c1];
                        for (int c2 = 0; c2 < this.yyParents[y].length; c2++)
                            auxM[this.yxParents[y][c1]] -= this.qEInvPsi[y] *
                                    this.rawData[i][this.yyParents[y][c2]] *
                                    this.qELambdaLambda[d][y][c2][this.yyParents[y].length + c1];
                        auxS[this.yxParents[y][c1]][this.yxParents[y][c1]] += this.qEInvPsi[y] *
                                this.qELambdaLambda[d][y][this.yyParents[y].length + c1][this.yyParents[y].length + c1];
                        for (int c2 = c1 + 1; c2 < this.yxParents[y].length; c2++) {
                            double aux = this.qEInvPsi[y] *
                                         this.qELambdaLambda[d][y]
                                                            [this.yyParents[y].length + c1][this.yyParents[y].length + c2];
                            if (this.yxParents[y][c1] < this.yxParents[y][c2])
                                auxS[this.yxParents[y][c1]][this.yxParents[y][c2]] += aux;
                            else
                                auxS[this.yxParents[y][c2]][this.yxParents[y][c1]] += aux;
                        }
                    }
                }
                for (int v1 = 0; v1 < this.numLatents; v1++) {
                    auxM[v1] += priorIntercept[v1];
                    for (int v2 = v1; v2 < this.numLatents; v2++)
                         auxS[v1][v2] += this.xInvCovPrior[d][v1][v2];
                }
                for (int v1 = 0; v1 < this.numLatents - 1; v1++)
                     for (int v2 = v1 + 1; v2 < this.numLatents; v2++)
                         auxS[v2][v1] = auxS[v1][v2];
                double auxSInv[][] = MatrixUtils.inverse(auxS);
                this.qXMean[d][i] = MatrixUtils.product(auxSInv, auxM);
                this.qEX[d][i] = this.qXMean[d][i];
                this.qXCov[d][i] = auxSInv;
                for (int x1 = 0; x1 < this.numLatents; x1++)
                    for (int x2 = x1; x2 < this.numLatents; x2++)
                        this.qEXX[d][i][x1][x2] = this.qEXX[d][i][x2][x1] = this.qXCov[d][i][x1][x2] +
                                                                            this.qEX[d][i][x1] * this.qEX[d][i][x2];
                for (int x1 = 0; x1 < this.numMainLatents; x1++)
                    for (int x2 = 0; x2 < this.numMainLatents; x2++)
                        this.qMainXCov[d][i][x1][x2] = this.qXCov[d][i][x1][x2];
            }
        }
        computeLikelihoodKernels();
    }

    protected void computeNewLikelihoodKernels(int y) {
        for (int d = 0; d < this.numMixtures; d++)
            for (int i = 0; i < this.sampleSize; i++) {
                double varYOthers = 0.;
                for (int y2 = 0; y2 < this.newYYParents.length; y2++)
                    varYOthers += this.newQELambda[d][y2] * this.rawData[i][y] * this.rawData[i][this.newYYParents[y2]];
                for (int x = 0; x < this.newYXParents.length; x++)
                    varYOthers += this.newQELambda[d][this.newYYParents.length + x] *
                                  this.rawData[i][y] * this.qEX[d][i][this.yxParents[y][x]];
                this.newLikelihoodKernels[d][i] = this.rawData[i][y] * this.rawData[i][y]  - 2. * varYOthers;

                double varYpOthers = 0.;
                for (int y2 = 0; y2 < this.newYYParents.length; y2++)
                    for (int y3 = 0; y3 < this.newYYParents.length; y3++)
                         varYpOthers += this.newQELambdaLambda[d][y2][y3] *
                                        this.rawData[i][this.newYYParents[y2]] *
                                        this.rawData[i][this.newYYParents[y3]];
                for (int y2 = 0; y2 < this.newYYParents.length; y2++)
                    for (int x = 0; x < this.newYXParents.length; x++)
                         varYpOthers += 2 * this.newQELambdaLambda[d][y2][this.newYYParents.length + x] *
                                        this.rawData[i][this.newYYParents[y2]] *
                                        this.qEX[d][i][this.newYXParents[x]];
                this.newLikelihoodKernels[d][i] += varYpOthers;

                double varXpOthers = 0.;
                for (int x = 0; x < this.newYXParents.length; x++) {
                    for (int x2 = 0; x2 < this.newYXParents.length; x2++)
                         varXpOthers += this.newQELambdaLambda[d][this.newYYParents.length + x][this.newYYParents.length + x2] *
                                        this.qEXX[d][i][this.newYXParents[x]][this.newYXParents[x2]];
                }
                this.newLikelihoodKernels[d][i] += varXpOthers;
            }
    }

    public Object clone() {
        VariationalMimMixture output = new VariationalMimMixture(this.dataSet, this.originalGraph,
                                                                 this.freeVariables, this.numMixtures);
        backupParams(output);
        return output;
    }

    //========================================================

    public double fastBidirectedEdgeAdditionEval(Node node1, Node node2) {
        return fastBidirectedEdgeAddition(node1, node2, false);
    }

    public double fastBidirectedEdgeAdditionEvalCommit(Node node1, Node node2) {
        return fastBidirectedEdgeAddition(node1, node2, true);
    }

    private double fastBidirectedEdgeAddition(Node node1, Node node2, boolean commit) {
        List variables = new ArrayList(); variables.add(node1); variables.add(node2);
        double currentScore = this.score;

        int ys[] = new int[variables.size()];
        for (int v = 0; v < variables.size(); v++) {
            ys[v] = -1;
            for (int i = 0; i < this.numVars; i++) {
                if (this.indicators[i].toString().equals(variables.get(v).toString())) {
                    ys[v] = i;
                    break;
                }
            }
            if (ys[v] < 0)
                throw new RuntimeException("This variable does not exist: " + variables.get(v));
        }
        //Remove its penalty contribution
        for (int d = 0; d < this.numMixtures; d++)
            for (int v = 0; v < variables.size(); v++) {
                //Remove more of its penalty contribution
                currentScore += RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][ys[v]], this.qLambdaCov[d][ys[v]],
                                                             this.lambdaMeanPrior[d][ys[v]], this.lambdaCovPrior[d][ys[v]]);
                //Remove its likelihood contribution
                for (int i = 0; i < this.sampleSize; i++)
                    currentScore += this.qS[d][i] * 0.5 * this.qEInvPsi[ys[v]] * this.likelihoodKernels[d][i][ys[v]];
            }
        //Remove more of its likelihood contribution
        for (int v = 0; v < variables.size(); v++)
            currentScore -= 0.5 * Math.log(this.qEInvPsi[ys[v]]) * this.sampleSize;
        //System.out.println("Score difference = " + (this.score - currentScore));
        //Prepare parameters: observed
        double backupInvCovLambdaPrior[] = new double[this.numVars];
        for (int i = 0; i < this.numVars; i++)
             backupInvCovLambdaPrior[i] = this.lambdaInvCovPriorParam[i];
        double backupQLambdaMean[][][] = new double[this.numMixtures][2][];
        double backupQLambdaCov[][][][] = new double[this.numMixtures][2][][];
        double backupQELambda[][][] = new double[this.numMixtures][2][];
        double backupQELambdaLambda[][][][] = new double[this.numMixtures][2][][];
        double backupQEInvPsi[] = new double[2];
        double backupLambdaMeanPrior[][][] = new double[this.numMixtures][2][];
        double backupLambdaInvCovPrior[][][][] = new double[this.numMixtures][2][][];
        double backupLambdaCovPrior[][][][] = new double[this.numMixtures][2][][];
        double backupLikelihoodKernels[][][] = new double[this.numMixtures][this.sampleSize][2];

        Random r = new Random();
        for (int i = 0; i < this.numVars; i++)
            this.lambdaInvCovPriorParam[i] = r.nextDouble();
        for (int i = 0; i < 2; i++) {
            backupQEInvPsi[i] = this.qEInvPsi[ys[i]];
            this.qEInvPsi[ys[i]] = r.nextDouble();
            for (int d = 0; d < this.numMixtures; d++) {
                backupLambdaMeanPrior[d][i] = this.lambdaMeanPrior[d][ys[i]];
                this.lambdaMeanPrior[d][ys[i]] = new double[this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1];
                for (int p = 0; p < this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1; p++)
                     this.lambdaMeanPrior[d][ys[i]][p] = 0.;
                backupLambdaInvCovPrior[d][i] = this.lambdaInvCovPrior[d][ys[i]];
                this.lambdaInvCovPrior[d][ys[i]] = new double[this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1]
                                                             [this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1];
                backupLambdaCovPrior[d][i] = this.lambdaCovPrior[d][ys[i]];
                this.lambdaCovPrior[d][ys[i]] = new double[this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1]
                                                          [this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1];
                for (int p1 = 0; p1 < this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1; p1++)
                    for (int p2 = 0; p2 < this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1; p2++) {
                        if (p1 == p2) {
                            this.lambdaInvCovPrior[d][ys[i]][p1][p2] = this.lambdaInvCovPriorParam[p1];
                            this.lambdaCovPrior[d][ys[i]][p1][p2] = 1. / this.lambdaInvCovPriorParam[p1];
                        }
                        else {
                            this.lambdaInvCovPrior[d][ys[i]][p1][p2] = 0.;
                            this.lambdaCovPrior[d][ys[i]][p1][p2] = 0.;
                        }
                    }
                backupQLambdaMean[d][i] = this.qLambdaMean[d][ys[i]];
                this.qLambdaMean[d][ys[i]] = new double[this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1];
                backupQELambda[d][i] = this.qELambda[d][ys[i]];
                this.qELambda[d][ys[i]] = new double[this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1];
                for (int p = 0; p < this.qLambdaMean[d][ys[i]].length; p++)
                    this.qLambdaMean[d][ys[i]][p] = this.qELambda[d][ys[i]][p] = r.nextDouble();
                backupQLambdaCov[d][i] = this.qLambdaCov[d][ys[i]];
                this.qLambdaCov[d][ys[i]] = new double[this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1]
                                                      [this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1];
                backupQELambdaLambda[d][i] = this.qELambdaLambda[d][ys[i]];
                this.qELambdaLambda[d][ys[i]] = new double[this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1]
                                                          [this.yyParents[ys[i]].length + this.yxParents[ys[i]].length + 1];
                for (int p1 = 0; p1 < this.qLambdaCov[d][ys[i]].length; p1++)
                    for (int p2 = 0; p2 < this.qLambdaCov[d][ys[i]].length; p2++) {
                        if (p1 == p2)
                            this.qLambdaCov[d][ys[i]][p1][p2] = this.qELambdaLambda[d][ys[i]][p1][p2] = 1.;
                        else
                            this.qLambdaCov[d][ys[i]][p1][p2] = this.qELambdaLambda[d][ys[i]][p1][p2] = 0.;
                    }
                for (int s = 0; s < this.sampleSize; s++)
                    backupLikelihoodKernels[d][s][i] = this.likelihoodKernels[d][s][ys[i]];
            }
        }

        //Prepare parameters: latent
        Node    backupLatents[] = this.latents;
        boolean backupIsConfounder[] = this.isConfounder;
        Edge    backupConfounders[] = this.confounders;
        Graph   backupGraph = this.graph;
        int     backupXXParents[][] = this.xxParents;
        int     backupXYParents[][] = this.xyParents;
        double  backupBetaMuInvCovPrior[][][][] = this.betaMuInvCovPrior;
        double  backupBetaMuCovPrior[][][][] = this.betaMuCovPrior;
        double  backupBetaMuMeanPrior[][][] = this.betaMuMeanPrior;
        double  backupXInvCovPrior[][][] = this.xInvCovPrior;
        double  backupXMeanPrior[][] = this.xMeanPrior;
        double  backupQXMean[][][] = this.qXMean, backupQXCov[][][][] = this.qXCov;
        double  backupQEX[][][] = this.qEX, backupQEXX[][][][] = this.qEXX;
        double  backupQBetaMuMean[][][] = this.qBetaMuMean, backupQBetaMuCov[][][][] = this.qBetaMuCov;
        double  backupQEBM[][][] = this.qEBM, backupQEBMBM[][][][] = this.qEBMBM;
        double  backupUpsilon[][] = this.upsilon;

        this.latents = new Node[this.latents.length + 1];
        for (int i = 0; i < backupLatents.length; i++)
             this.latents[i] = backupLatents[i];
        this.latents[this.latents.length - 1] = new GraphNode("__L " + this.latents.length);
        this.isConfounder = new boolean[this.isConfounder.length + 1];
        for (int i = 0; i < backupIsConfounder.length; i++)
             this.isConfounder[i] = backupIsConfounder[i];
        this.isConfounder[this.isConfounder.length - 1] = true;
        this.numConfounders++; this.numLatents++;
        this.confounders = new Edge[this.confounders.length + 1];
        for (int i = 0; i < backupConfounders.length; i++)
             this.confounders[i] = backupConfounders[i];
        Edge newEdge = new Edge(node1, node2, Endpoint.ARROW, Endpoint.ARROW);
        this.confounders[this.confounders.length - 1] = newEdge;
        this.graph = new EdgeListGraph(this.graph);
        this.graph.addEdge(newEdge);
        this.xxParents = new int[this.latents.length][];
        for (int i = 0; i < this.latents.length - 1; i++)
            this.xxParents[i] = backupXXParents[i];
        this.xxParents[this.latents.length - 1] = new int[0];
        this.xyParents = new int[this.latents.length][];
        for (int i = 0; i < this.latents.length - 1; i++)
            this.xyParents[i] = backupXYParents[i];
        this.xyParents[this.latents.length - 1] = new int[0];
        this.betaMuInvCovPrior = new double[this.numMixtures][this.numLatents][][];
        this.betaMuCovPrior = new double[this.numMixtures][this.numLatents][][];
        this.betaMuMeanPrior  = new double[this.numMixtures][this.numLatents][];
        for (int d = 0; d < this.numMixtures; d++) {
            for (int l = 0; l < this.numLatents - 1; l++) {
                this.betaMuInvCovPrior[d][l] = backupBetaMuInvCovPrior[d][l];
                this.betaMuCovPrior[d][l] = backupBetaMuCovPrior[d][l];
                this.betaMuMeanPrior[d][l] = backupBetaMuMeanPrior[d][l];
            }
            this.betaMuInvCovPrior[d][this.numLatents - 1] = new double[0][0];
            this.betaMuCovPrior[d][this.numLatents - 1] = new double[0][0];
            this.betaMuMeanPrior[d][this.numLatents - 1] = new double[0];
        }
        this.xInvCovPrior = new double[this.numMixtures][this.numLatents][this.numLatents];
        this.xMeanPrior = new double[this.numMixtures][this.numLatents];
        for (int d = 0; d < this.numMixtures; d++) {
            for (int l1 = 0; l1 < this.numLatents - 1; l1++) {
                this.xMeanPrior[d][l1] = backupXMeanPrior[d][l1];
                this.xInvCovPrior[d][l1][this.numLatents - 1] = this.xInvCovPrior[d][this.numLatents - 1][l1] = 0.;
                for (int l2 = l1; l2 < this.numLatents - 1; l2++)
                    this.xInvCovPrior[d][l1][l2] = this.xInvCovPrior[d][l2][l1] = backupXInvCovPrior[d][l1][l2];
            }
            this.xMeanPrior[d][this.numLatents - 1] = 0.;
            this.xInvCovPrior[d][this.numLatents - 1][this.numLatents - 1] = 1.;
        }
        this.qXMean = new double[this.numMixtures][this.sampleSize][this.numLatents];
        this.qXCov = new double[this.numMixtures][this.sampleSize][this.numLatents][this.numLatents];
        this.qEX = new double[this.numMixtures][this.sampleSize][this.numLatents];
        this.qEXX = new double[this.numMixtures][this.sampleSize][this.numLatents][this.numLatents];
        for (int d = 0; d < this.numMixtures; d++) {
            for (int i = 0; i < this.sampleSize; i++) {
                for (int l1 = 0; l1 < this.numLatents - 1; l1++) {
                    this.qXMean[d][i][l1] = this.qEX[d][i][l1] = backupQXMean[d][i][l1];
                    for (int l2 = l1; l2 < this.numLatents - 1; l2++) {
                        this.qEXX[d][i][l1][l2] = this.qEXX[d][i][l2][l1] = backupQEXX[d][i][l1][l2];
                        this.qXCov[d][i][l1][l2] = this.qXCov[d][i][l2][l1] = backupQXCov[d][i][l1][l2];
                    }
                }
                this.qXCov[d][i][this.numLatents - 1][this.numLatents - 1] = 1.;
                this.qEX[d][i][this.numLatents - 1] = this.qXMean[d][i][this.numLatents - 1] = r.nextGaussian();
                this.qEXX[d][i][this.numLatents - 1][this.numLatents - 1] =
                        this.qXCov[d][i][this.numLatents - 1][this.numLatents - 1] +
                        this.qEX[d][i][this.numLatents - 1] * this.qEX[d][i][this.numLatents - 1];
                for (int l = 0; l < this.numLatents - 1; l++)
                    this.qEXX[d][i][l][this.numLatents - 1] = this.qEXX[d][i][this.numLatents - 1][l] =
                            this.qEX[d][i][l] * this.qEX[d][i][this.numLatents - 1];
            }
        }
        this.qBetaMuMean = new double[this.numMixtures][this.numLatents][];
        this.qBetaMuCov = new double[this.numMixtures][this.numLatents][][];
        this.qEBM = new double[this.numMixtures][this.numLatents][];
        this.qEBMBM = new double[this.numMixtures][this.numLatents][][];
        for (int d = 0; d < this.numMixtures; d++) {
            for (int l = 0; l < this.numLatents - 1; l++) {
                this.qBetaMuMean[d][l] = backupQBetaMuMean[d][l];
                this.qEBM[d][l] = backupQEBM[d][l];
                this.qBetaMuCov[d][l] = backupQBetaMuCov[d][l];
                this.qEBMBM[d][l] = backupQEBMBM[d][l];
            }
            this.qBetaMuMean[d][this.numLatents - 1] = new double[0];
            this.qEBM[d][this.numLatents - 1] = new double[0];
            this.qBetaMuCov[d][this.numLatents - 1] = new double[0][0];
            this.qEBMBM[d][this.numLatents - 1] = new double[0][0];
        }
        this.upsilon = new double[this.numMixtures][this.numLatents];
        for (int d = 0; d < this.numMixtures; d++) {
            for (int l = 0; l < this.numLatents - 1; l++)
                this.upsilon[d][l] = backupUpsilon[d][l];
            this.upsilon[d][this.numLatents - 1] = 1.;
        }

        int backupYXParents[][] = new int[2][];
        for (int v = 0; v < variables.size(); v++) {
            backupYXParents[v] = this.yxParents[ys[v]];
            this.yxParents[ys[v]] = new int[this.yxParents[ys[v]].length + 1];
            for (int i = 0; i < this.yxParents[ys[v]].length - 1; i++)
                 this.yxParents[ys[v]][i] = backupYXParents[v][i];
            this.yxParents[ys[v]][this.yxParents[ys[v]].length - 1] = this.numLatents - 1;
        }

        //Finally, do the actual optimization
        currentScore += fastOptimizeCumLatent(ys);
        if (commit) {
            this.score = currentScore;
        } else {
            for (int v = 0; v < variables.size(); v++)
                 this.yxParents[ys[v]] = backupYXParents[v];
            for (int i = 0; i < this.numVars; i++)
                this.lambdaInvCovPriorParam[i] = backupInvCovLambdaPrior[i];
            for (int i = 0; i < 2; i++) {
                for (int d = 0; d < this.numMixtures; d++) {
                    this.qLambdaMean[d][ys[i]] = backupQLambdaMean[d][i];
                    this.qLambdaCov[d][ys[i]] = backupQLambdaCov[d][i];
                    this.qELambda[d][ys[i]] = backupQELambda[d][i];
                    this.qELambdaLambda[d][ys[i]] = backupQELambdaLambda[d][i];
                    this.lambdaMeanPrior[d][ys[i]] = backupLambdaMeanPrior[d][i];
                    this.lambdaInvCovPrior[d][ys[i]] = backupLambdaInvCovPrior[d][i];
                    this.lambdaCovPrior[d][ys[i]] = backupLambdaCovPrior[d][i];
                    for (int s = 0; s < this.sampleSize; s++)
                        this.likelihoodKernels[d][s][ys[i]] = backupLikelihoodKernels[d][s][i];
                }
                this.qEInvPsi[ys[i]] = backupQEInvPsi[i];
            }
            this.numConfounders--; this.numLatents--;
            this.latents = backupLatents;
            this.isConfounder = backupIsConfounder;
            this.confounders = backupConfounders;
            this.graph = backupGraph;
            this.xxParents = backupXXParents;
            this.xyParents = backupXYParents;
            this.betaMuInvCovPrior = backupBetaMuInvCovPrior;
            this.betaMuCovPrior = backupBetaMuCovPrior;
            this.betaMuMeanPrior = backupBetaMuMeanPrior;
            this.xInvCovPrior = backupXInvCovPrior;
            this.xMeanPrior = backupXMeanPrior;
            this.qXMean = backupQXMean; this.qXCov = backupQXCov;
            this.qEX = backupQEX; this.qEXX = backupQEXX;
            this.qBetaMuMean = backupQBetaMuMean; this.qBetaMuCov = backupQBetaMuCov;
            this.qEBM = backupQEBM; this.qEBMBM = backupQEBMBM;
            this.upsilon = backupUpsilon;
        }
        return currentScore;
    }

    protected void updateSingleQXConfounder(int x, int ys[]) {
        int xs[] = new int[ys.length];
        for (int v = 0; v < ys.length; v++)
            for (int p = 0; p < this.yxParents[ys[v]].length; p++) {
                if (this.yxParents[ys[v]][p] == x) {
                    xs[v] = p;
                    break;
                }
            }
        double auxS, auxM;
        for (int d = 0; d < this.numMixtures; d++) {
            for (int i = 0; i < this.sampleSize; i++) {
                computeSingleConfounderPX(i, d, x);
                auxM = 0.;
                auxS = 0.;
                for (int y = 0; y < ys.length; y++) {
                    auxM += this.qEInvPsi[ys[y]] * this.rawData[i][ys[y]] *
                            this.qELambda[d][ys[y]][this.yyParents[ys[y]].length + xs[y]];
                    for (int c2 = 0; c2 < this.yyParents[ys[y]].length; c2++)
                        auxM -= this.qEInvPsi[ys[y]] *
                                this.rawData[i][this.yyParents[ys[y]][c2]] *
                                this.qELambdaLambda[d][ys[y]][c2][this.yyParents[ys[y]].length + xs[y]];
                    for (int c2 = 0; c2 < this.yxParents[ys[y]].length; c2++) {
                        if (c2 == xs[y])
                            continue;
                        auxM -= this.qEInvPsi[ys[y]] *
                                this.qEX[d][i][this.yxParents[ys[y]][c2]] *
                                this.qELambdaLambda[d][ys[y]][this.yyParents[ys[y]].length + c2][this.yyParents[ys[y]].length + xs[y]];
                    }
                    auxS += this.qEInvPsi[ys[y]] *
                            this.qELambdaLambda[d][ys[y]][this.yyParents[ys[y]].length + xs[y]][this.yyParents[ys[y]].length + xs[y]];
                }
                auxS += this.xInvCovPrior[d][x][x];
                double auxSInv = 1. / auxS;
                this.qXMean[d][i][x] = auxSInv * auxM;
                this.qEX[d][i][x] = this.qXMean[d][i][x];
                this.qXCov[d][i][x][x] = auxSInv;
                this.qEXX[d][i][x][x] = this.qXCov[d][i][x][x] + this.qEX[d][i][x] * this.qEX[d][i][x];
                for (int l1 = 0; l1 < this.numLatents - 1; l1++) {
                    this.qEXX[d][i][x][l1] = this.qEXX[d][i][l1][x] = this.qEX[d][i][l1] * this.qEX[d][i][x];
                    this.qXCov[d][i][x][l1] = this.qXCov[d][i][l1][x] = 0.;
                }
            }
        }
        computeMultipleLikelihoodKernels(ys);
    }

    protected void computeSingleConfounderPX(int i, int d, int x) {
        this.xInvCovPrior[d][x][x] = this.upsilon[d][x];
        this.xMeanPrior[d][x] = 0.;
    }

    public double fastOptimizeCumLatent(int ys[]) {
        double f, newF = newMultipleFEval(ys);
        //System.out.println(" ------ F0 = " + newF);
        int iter = 0;
        do {
            f = newF;
            updateMultipleQLambda(ys); updateLambdaHyperParameters();
            updateSingleQXConfounder(this.numLatents - 1, ys);
            updateMultipleQPsi(ys);
            newF = newMultipleFEval(ys);
            //System.out.println(" ------ F = " + newF);
            iter++;
        } while (newF - f > MIN_TOLERANCE && iter < 5);
        if (this.converging == false)
            System.exit(0);
        return newF;
    }

    private double newMultipleFEval(int ys[]) {
        double f = 0.;
        for (int d = 0; d < this.numMixtures; d++) {
            for (int y = 0; y < ys.length; y++) {
                f -= RBASProbUtils.klDivergenceNormal(this.qLambdaMean[d][ys[y]], this.qLambdaCov[d][ys[y]],
                                                  this.lambdaMeanPrior[d][ys[y]], this.lambdaCovPrior[d][ys[y]]);
                for (int i = 0; i < this.sampleSize; i++)
                    f -= 0.5 * this.qS[d][i] * this.qEInvPsi[ys[y]]* this.likelihoodKernels[d][i][ys[y]];
            }
        }
        for (int i = 0; i < ys.length; i++)
            f += 0.5 * Math.log(this.qEInvPsi[ys[i]]) * this.sampleSize;
        return f;
    }

    protected void updateMultipleQLambda(int ys[]) {
        for (int d = 0; d < this.numMixtures; d++)
            for (int v = 0; v < ys.length; v++) {
                int totalParents = this.yyParents[ys[v]].length + this.yxParents[ys[v]].length;
                double auxS[][] = new double[totalParents][totalParents];
                double auxMean[] = new double[totalParents];
                for (int i = 0; i < auxS.length; i++) {
                    auxMean[i] = 0.;
                    for (int j = 0; j < auxS.length; j++)
                        auxS[i][j] = 0.;
                }
                for (int i = 0; i < this.sampleSize; i++) {
                    double constant = this.qS[d][i] * this.qEInvPsi[ys[v]], nextTerm;
                    for (int p1 = 0; p1 < this.yyParents[ys[v]].length; p1++) {
                        for (int p2 = 0; p2 < this.yyParents[ys[v]].length; p2++)
                            auxS[p1][p2] += constant *
                                            this.rawData[i][this.yyParents[ys[v]][p1]] * this.rawData[i][this.yyParents[ys[v]][p2]];
                        for (int p2 = 0; p2 < this.yxParents[ys[v]].length; p2++) {
                            nextTerm = constant *
                                       this.rawData[i][this.yyParents[ys[v]][p1]] * this.qEX[d][i][this.yxParents[ys[v]][p2]];
                            auxS[p1][this.yyParents[ys[v]].length + p2] += nextTerm;
                            auxS[this.yyParents[ys[v]].length + p2][p1] += nextTerm;
                        }
                        auxMean[p1] += constant * this.rawData[i][ys[v]] * this.rawData[i][this.yyParents[ys[v]][p1]];
                    }
                    for (int p1 = 0; p1 < this.yxParents[ys[v]].length; p1++) {
                        for (int p2 = 0; p2 < this.yxParents[ys[v]].length; p2++)
                            auxS[this.yyParents[ys[v]].length + p1][this.yyParents[ys[v]].length + p2] += constant *
                                    this.qEXX[d][i][this.yxParents[ys[v]][p1]][this.yxParents[ys[v]][p2]];
                        auxMean[this.yyParents[ys[v]].length + p1] +=
                                constant * this.rawData[i][ys[v]] * this.qEX[d][i][this.yxParents[ys[v]][p1]];
                    }
                }
                for (int t = 0; t < auxS.length; t++)
                     auxS[t][t] += this.lambdaInvCovPrior[d][ys[v]][t][t];
                this.qLambdaCov[d][ys[v]] = MatrixUtils.inverse(auxS);
                this.qLambdaMean[d][ys[v]] = MatrixUtils.product(this.qLambdaCov[d][ys[v]], auxMean);
                this.qELambda[d][ys[v]] = this.qLambdaMean[d][ys[v]];
                //System.out.println(this.qELambda[d][ys[v]][1]);
                for (int v1 = 0; v1 < totalParents; v1++)
                    for (int v2 = 0; v2 < totalParents; v2++)
                        this.qELambdaLambda[d][ys[v]][v1][v2] = this.qLambdaCov[d][ys[v]][v1][v2] +
                                                            this.qLambdaMean[d][ys[v]][v1] * this.qLambdaMean[d][ys[v]][v2];
            }
        computeMultipleLikelihoodKernels(ys);
    }

    protected void computeMultipleLikelihoodKernels(int ys[]) {
        mainLoop:
        for (int d = 0; d < this.numMixtures; d++)
            for (int i = 0; i < this.sampleSize; i++)
                for (int y = 0; y < ys.length; y++) {

                    double varYOthers = 0.;
                    for (int y2 = 0; y2 < this.yyParents[ys[y]].length; y2++)
                        varYOthers += this.qELambda[d][ys[y]][y2] * this.rawData[i][ys[y]] * this.rawData[i][this.yyParents[ys[y]][y2]];
                    for (int x = 0; x < this.yxParents[ys[y]].length; x++)
                        varYOthers += this.qELambda[d][ys[y]][this.yyParents[ys[y]].length + x] *
                                      this.rawData[i][ys[y]] * this.qEX[d][i][this.yxParents[ys[y]][x]];
                    this.likelihoodKernels[d][i][ys[y]] = this.rawData[i][ys[y]] * this.rawData[i][ys[y]]  - 2. * varYOthers;

                    double varYpOthers = 0.;
                    for (int y2 = 0; y2 < this.yyParents[ys[y]].length; y2++)
                        for (int y3 = 0; y3 < this.yyParents[ys[y]].length; y3++)
                             varYpOthers += this.qELambdaLambda[d][ys[y]][y2][y3] *
                                            this.rawData[i][this.yyParents[ys[y]][y2]] *
                                            this.rawData[i][this.yyParents[ys[y]][y3]];
                    for (int y2 = 0; y2 < this.yyParents[ys[y]].length; y2++)
                        for (int x = 0; x < this.yxParents[ys[y]].length; x++)
                             varYpOthers += 2 * this.qELambdaLambda[d][ys[y]][y2][this.yyParents[ys[y]].length + x] *
                                            this.rawData[i][this.yyParents[ys[y]][y2]] *
                                            this.qEX[d][i][this.yxParents[ys[y]][x]];
                    this.likelihoodKernels[d][i][ys[y]] += varYpOthers;

                    double varXpOthers = 0.;
                    for (int x = 0; x < this.yxParents[ys[y]].length; x++) {
                        for (int x2 = 0; x2 < this.yxParents[ys[y]].length; x2++)
                             varXpOthers += this.qELambdaLambda[d][ys[y]][this.yyParents[ys[y]].length + x][this.yyParents[ys[y]].length + x2] *
                                            this.qEXX[d][i][this.yxParents[ys[y]][x]][this.yxParents[ys[y]][x2]];
                    }
                    this.likelihoodKernels[d][i][ys[y]] += varXpOthers;

                    if (this.likelihoodKernels[d][i][ys[y]] < 0.) {
                        this.converging = false;
                        System.out.println("No convergence");
                        break mainLoop;
                    }
                }
    }

    protected void updateMultipleQPsi(int ys[]) {
        for (int y = 0; y < ys.length; y++) {
            double dataContrib = 0.;
            for (int d = 0; d < this.numMixtures; d++)
                for (int i = 0; i < this.sampleSize; i++)
                    dataContrib += this.qS[d][i] * this.likelihoodKernels[d][i][ys[y]];
            if (dataContrib < 0.) {
                this.converging = false;
                return;
            }
            this.qEInvPsi[ys[y]] = this.sampleSize / dataContrib;
        }
    }

    public void addDirectedEdge(Node node1, Node node2) {
        List variables = new ArrayList(); variables.add(node1); variables.add(node2);
        int ys[] = new int[2];
        for (int v = 0; v < 2; v++) {
            ys[v] = -1;
            if (((Node) variables.get(v)).getNodeType() == NodeType.LATENT) {
                for (int i = 0; i < this.numVars; i++)
                    if (this.latents[i].toString().equals(((Node) variables.get(v)).getNodeType().toString())) {
                        ys[v] = i;
                        break;
                    }
            } else {
                for (int i = 0; i < this.numVars; i++)
                    if (this.indicators[i].toString().equals(((Node) variables.get(v)).getNodeType().toString())) {
                        ys[v] = i;
                        break;
                    }
            }
            if (ys[v] < 0)
                throw new RuntimeException("Node not found");
        }
        if (node1.getNodeType() == NodeType.LATENT) {
            if (node2.getNodeType() == NodeType.LATENT) {
                int backupXXParents[] = this.xxParents[ys[1]];
                this.xxParents[ys[1]] = new int[backupXXParents.length + 1];
                for (int i = 0; i < backupXXParents.length; i++)
                    this.xxParents[ys[1]][i] = backupXXParents[i];
                this.xxParents[ys[1]][backupXXParents.length] = ys[1];
                for (int d = 0; d < this.numMixtures; d++) {
                    this.betaMuInvCovPrior[d][ys[1]] = new double[this.xxParents[ys[1]].length][this.xxParents[ys[1]].length];
                    this.betaMuCovPrior[d][ys[1]] = new double[this.xxParents[ys[1]].length][this.xxParents[ys[1]].length];
                    this.betaMuMeanPrior[d][ys[1]] = new double[this.xxParents[ys[1]].length];
                    int numParents = this.xxParents[ys[1]].length + this.xyParents[ys[1]].length;
                    for (int v1 = 0; v1 < numParents; v1++) {
                        this.betaMuMeanPrior[d][ys[1]][v1] = 0.; //This *has* to be zero.
                        for (int v2 = 0; v2 < numParents; v2++) {
                            if (v1 == v2) {
                                this.betaMuInvCovPrior[d][ys[1]][v1][v2] = this.betaMuInvCovPriorParam;
                                this.betaMuCovPrior[d][ys[1]][v1][v2] = 1. / this.betaMuInvCovPriorParam;
                            }
                            else {
                                this.betaMuCovPrior[d][ys[1]][v1][v2] = 0.;
                                this.betaMuInvCovPrior[d][ys[1]][v1][v2] = 0.;
                            }
                        }
                    }
                }
            }
        }
    }
}



